#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import numpy as np
import pyomo.environ as pyo
import warnings
import copy
import time

from abc import abstractmethod, ABC
from idaes.core.solvers import get_solver

from idaes.core.surrogate.pysmo import sampling
from pyomo.common.deprecation import deprecation_warning
from pyomo.common.config import ConfigValue
from pyomo.common.modeling import unique_component_name
from pyomo.core.base import _VarData, _ExpressionData
from pyomo.core.base.param import _ParamData
from pyomo.common.dependencies import attempt_import

requests, requests_available = attempt_import("requests")

from watertap.tools.parameter_sweep.parameter_sweep_writer import ParameterSweepWriter
from watertap.tools.parameter_sweep.sampling_types import SamplingType, LinearSample

from watertap.tools.parallel.parallel_manager_factory import create_parallel_manager

from watertap.tools.parameter_sweep.model_manager import ModelManager
from watertap.tools.parameter_sweep.paramter_sweep_parallel_utils import (
    _ParameterSweepParallelUtils,
    return_none,
)


def _default_optimize(model, options=None, tee=False):
    """
    Default optimization function used in parameter_sweep.
    Optimizes ``model`` using the IDAES default solver.
    Raises a RuntimeError if the TerminationCondition is not optimal

    Arguments:

        model : A Pyomo ConcreteModel to optimize

        options (optional) : Solver options to pass into idaes.core.utils.get_solver.
                             Default is None
        tee (options) : To display the solver log. Default it False

    """
    solver = get_solver(options=options)
    results = solver.solve(model, tee=tee)
    return results


class _ParameterSweepBase(ABC):
    CONFIG = ParameterSweepWriter.CONFIG()

    CONFIG.declare(
        "build_model",
        ConfigValue(
            default=None,
            # domain=function,
            description="Function for building the model.",
        ),
    )

    CONFIG.declare(
        "build_model_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword argument for the model build function for the parameter sweep.",
        ),
    )
    CONFIG.declare(
        "build_sweep_params",
        ConfigValue(
            default=None,
            # domain=function,
            description="Function for building the sweep_paramters",
        ),
    )
    CONFIG.declare(
        "build_sweep_params_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword argument for the build sweep params function for the parameter sweep.",
        ),
    )

    CONFIG.declare(
        "build_outputs",
        ConfigValue(
            default=None,
            # domain=function,
            description="Function for building outputs",
        ),
    )
    CONFIG.declare(
        "build_outputs_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword argument for the build outputs function for the parameter sweep.",
        ),
    )

    CONFIG.declare(
        "optimize_function",
        ConfigValue(
            default=_default_optimize,
            # domain=function,
            description="Optimization function to be used for the parameter sweep.",
        ),
    )

    CONFIG.declare(
        "optimize_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword argument for the optimization function for the parameter sweep.",
        ),
    )
    CONFIG.declare(
        "initialize_function",
        ConfigValue(
            default=None,
            # domain=function,
            description="Function to reinitialize a flowsheet",
        ),
    )
    CONFIG.declare(
        "update_sweep_params_before_init",
        ConfigValue(
            default=False,
            # domain=function,
            description="Enables update of vars to sweep values before initilization (only enabled if init_before_sweep=True)",
        ),
    )
    CONFIG.declare(
        "initialize_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword arguments for the initialization function.",
        ),
    )

    CONFIG.declare(
        "initialize_before_sweep",
        ConfigValue(
            default=False,
            domain=bool,
            description="Initializing a model before every iteration.",
        ),
    )

    CONFIG.declare(
        "reinitialize_function",
        ConfigValue(
            default=None,
            # domain=function,
            description="Function to reinitialize a flowsheet",
        ),
    )
    CONFIG.declare(
        "reinitialize_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Keyword arguments for the reinitialization function.",
        ),
    )

    CONFIG.declare(
        "reinitialize_before_sweep",
        ConfigValue(
            default=False,
            domain=bool,
            description="Reinitializing a model before every iteration.",
        ),
    )

    CONFIG.declare(
        "probe_function",
        ConfigValue(
            default=None,
            description="Function to probe if a flowsheet configuration will work",
        ),
    )

    CONFIG.declare(
        "custom_do_param_sweep",
        ConfigValue(
            default=None,
            description="Alternative implementation of the parameter sweep function in case the user is doing unique analyses.",
        ),
    )

    CONFIG.declare(
        "custom_do_param_sweep_kwargs",
        ConfigValue(
            default=dict(),
            domain=dict,
            description="Alternative implementation of the parameter sweep function in case the user is doing unique analyses.",
        ),
    )

    CONFIG.declare(
        "publish_progress",
        ConfigValue(
            default=False,
            domain=bool,
            description="Boolean to decide whether information about how many iterations of the parameter sweep have completed should be sent.",
        ),
    )

    CONFIG.declare(
        "publish_address",
        ConfigValue(
            default="http://localhost:8888",
            domain=str,
            description="Address to which the parameter sweep progress will be sent.",
        ),
    )

    CONFIG.declare(
        "number_of_subprocesses",
        ConfigValue(
            default=1,
            domain=int,
            description="Number of processes to fan out to locally - ignored if running under MPI.",
        ),
    )
    CONFIG.declare(
        "parallel_back_end",
        ConfigValue(
            default="ConcurrentFutures",
            domain=str,
            description="Backend for parallelization, if not useing MPI",
        ),
    )
    CONFIG.declare(
        "log_model_states",
        ConfigValue(
            default=False,
            domain=bool,
            description="Enables loging of model states during serial execution",
        ),
    )
    CONFIG.declare(
        "index_global_combo_array",
        ConfigValue(
            default=False,
            domain=bool,
            description="Will add indexing to global_combo_array, primarily used with differential parameter sweep tool",
        ),
    )

    def __init__(
        self,
        **options,
    ):
        parallel_manager_class = options.pop("parallel_manager_class", None)
        self.model = None

        self.model_manager = None
        self.config = self.CONFIG(options)
        self.parallel_manager = create_parallel_manager(
            parallel_manager_class=parallel_manager_class,
            number_of_subprocesses=self.config.number_of_subprocesses,
            parallel_back_end=self.config.parallel_back_end,
        )

        # Initialize the writer
        self.writer = ParameterSweepWriter(
            self.parallel_manager,
            csv_results_file_name=self.config.csv_results_file_name,
            h5_results_file_name=self.config.h5_results_file_name,
            debugging_data_dir=self.config.debugging_data_dir,
            interpolate_nan_outputs=self.config.interpolate_nan_outputs,
            h5_parent_group_name=self.config.h5_parent_group_name,
        )

    @staticmethod
    def assign_variable_names(model, outputs):
        # Only assign output variable names to unassigned outputs
        exprs = pyo.Expression(pyo.Any)
        model.add_component(
            unique_component_name(model, "_parameter_sweep_expressions"), exprs
        )
        for output_name, _pyo_obj in outputs.items():
            if not isinstance(_pyo_obj, (_VarData, _ExpressionData, _ParamData)):
                # Add this object as an expression and assign a name
                exprs[output_name] = _pyo_obj
                outputs[output_name] = exprs[output_name]

    def _publish_updates(self, iteration, solve_status, solve_time):
        if not requests_available:
            raise ImportError(
                "requests (parameter_sweep optional dependency) not installed"
            )

        if self.config.publish_progress:
            publish_dict = {
                "worker_number": self.parallel_manager.get_rank(),
                "iteration": iteration,
                "solve_status": solve_status,
                "solve_time": solve_time,
            }

            return requests.put(self.config.publish_address, data=publish_dict)

    def _create_global_combo_array(self, d, sampling_type):
        num_var_params = len(d)
        param_values = []

        for k, v in d.items():
            # Build a vector of discrete values for this parameter
            p = v.sample()
            param_values.append(p)

        if sampling_type == SamplingType.FIXED:
            # Form an array with every possible combination of parameter values
            global_combo_array = np.array(np.meshgrid(*param_values, indexing="ij"))
            global_combo_array = global_combo_array.reshape(num_var_params, -1).T

        elif sampling_type == SamplingType.RANDOM:
            sorting = np.argsort(param_values[0])
            global_combo_array = np.vstack(param_values).T
            global_combo_array = global_combo_array[sorting, :]

        elif sampling_type == SamplingType.RANDOM_LHS:
            lb = [val[0] for val in param_values]
            ub = [val[1] for val in param_values]
            lhs = sampling.LatinHypercubeSampling(
                [lb, ub], number_of_samples=v.num_samples, sampling_type="creation"
            )
            global_combo_array = lhs.sample_points()
            sorting = np.argsort(global_combo_array[:, 0])
            global_combo_array = global_combo_array[sorting, :]

        else:
            raise ValueError(f"Unknown sampling type: {sampling_type}")

        # Test if the global_combo_array is in row-major order
        if not global_combo_array.flags.c_contiguous:
            # If not, return a copy of this array with row-major memory order
            global_combo_array = np.ascontiguousarray(global_combo_array)
        # add sample index for tracking of sample number in parallel schema
        if self.config.index_global_combo_array:
            sample_idx = np.arange(global_combo_array.shape[0]).reshape(-1, 1)
            global_combo_array = np.hstack((sample_idx, global_combo_array))
        return global_combo_array

    """
    Put together all of the parameter combinations that the sweep will be run for.
    """

    def _build_combinations(self, d, sampling_type, num_samples):
        # only build the full array of combinations on the root process. on the non-root
        # processes, initialize an empty array of the right size that will be synced
        # over from the root process.

        global_combo_array = self._create_global_combo_array(d, sampling_type)
        self.parallel_manager.sync_array_with_peers(global_combo_array)

        return global_combo_array

    def _divide_combinations(self, global_combo_array):
        # Split the total list of combinations into NUM_PROCS chunks,
        # one per each of the MPI ranks
        # divided_combo_array = np.array_split(global_combo_array, num_procs, axis=0)
        divided_combo_array = np.array_split(
            global_combo_array, self.parallel_manager.number_of_worker_processes()
        )

        # Return only this rank's portion of the total workload
        local_combo_array = divided_combo_array[self.parallel_manager.get_rank()]

        return local_combo_array

    def _get_object(self, model, pyomo_object):
        name = pyomo_object.name

        # seems to be a bug, as indexed var with [None] exists
        # but can't be found by find_component
        if "[None]" in name:
            name = name.replace("[None]", "")
            return model.find_component(name)[None]
        else:
            return model.find_component(name)

    def _update_model_values(self, m, param_dict, values):
        # remove index from values
        if self.config.index_global_combo_array:
            non_indexed_values = values[1:]
        else:
            non_indexed_values = values

        for k, item in enumerate(param_dict.values()):
            param = self._get_object(m, item.pyomo_object)
            if param.is_variable_type():
                # Fix the single value to values[k]
                param.fix(non_indexed_values[k])

            elif param.is_parameter_type():
                # Fix the single value to values[k]
                param.set_value(non_indexed_values[k])

            else:
                raise RuntimeError(f"Unrecognized Pyomo object {param}")

    def _aggregate_results_arr(self, global_results_dict, num_cases):
        global_results = np.zeros(
            (num_cases, len(global_results_dict["outputs"])), dtype=float
        )

        if self.parallel_manager.is_root_process():
            for i, (key, item) in enumerate(global_results_dict["outputs"].items()):
                global_results[:, i] = item["value"][:num_cases]

        self.parallel_manager.sync_array_with_peers(global_results)

        return global_results

    def _process_sweep_params(self, sweep_params):
        sampling_type = None

        # Check the list of parameters to make sure they are valid
        for k in sweep_params.keys():
            # Convert to using Sample class
            if isinstance(sweep_params[k], (list, tuple)):
                sweep_params[k] = LinearSample(*sweep_params[k])

            # Get the type of sampling
            current_sampling_type = sweep_params[k].sampling_type

            # Check to make sure only one sampling type is provided
            if sampling_type is None:
                sampling_type = current_sampling_type
            elif current_sampling_type != sampling_type:
                raise ValueError("Cannot mix sampling types")

        return sweep_params, sampling_type

    def _create_local_output_skeleton(self, model, sweep_params, outputs, num_samples):
        output_dict = {}
        output_dict["sweep_params"] = {}
        output_dict["outputs"] = {}

        # Store the inputs
        for param_name, sampling_obj in sweep_params.items():
            var = sampling_obj.pyomo_object
            output_dict["sweep_params"][
                param_name
            ] = self._create_component_output_skeleton(var, num_samples)

        if outputs is None:
            # No outputs are specified, so every Var, Expression, and Objective on the model should be saved
            for pyo_obj in model.component_data_objects(
                (pyo.Var, pyo.Expression, pyo.Objective, pyo.Param), active=True
            ):
                # We do however need to make sure that the short name for the inputs is used here
                for param_name, sampling_obj in sweep_params.items():
                    if pyo_obj.name == sampling_obj.pyomo_object.name:
                        output_dict["outputs"][
                            param_name
                        ] = self._create_component_output_skeleton(pyo_obj, num_samples)
                    else:
                        output_dict["outputs"][
                            pyo_obj.name
                        ] = self._create_component_output_skeleton(pyo_obj, num_samples)

        else:
            # Save only the outputs specified in the outputs dictionary
            for short_name, pyo_obj in outputs.items():
                output_dict["outputs"][
                    short_name
                ] = self._create_component_output_skeleton(
                    self._get_object(model, pyo_obj), num_samples
                )
        return output_dict

    def _create_component_output_skeleton(self, component, num_samples):
        comp_dict = {}
        comp_dict["value"] = np.zeros(num_samples, dtype=float)

        if hasattr(component, "lb"):
            comp_dict["lower bound"] = component.lb
        if hasattr(component, "ub"):
            comp_dict["upper bound"] = component.ub
        if hasattr(component, "get_units"):
            unit_obj = component.get_units()
            if unit_obj is not None:
                comp_dict["units"] = component.get_units().name
            else:
                comp_dict["units"] = "None"

        # Add information to this output that WILL NOT be written as part
        # of the file saving step.
        comp_dict["full_name"] = component.name

        return comp_dict

    def _update_local_output_dict(
        self, model, sweep_params, case_number, run_successful, output_dict
    ):
        # Get the inputs
        op_ps_dict = output_dict["sweep_params"]
        for key, item in sweep_params.items():
            # stores value actually applied to model, rather one assumed to be applied
            op_ps_dict[key]["value"][case_number] = self._get_object(
                model, item.pyomo_object
            ).value

        # Get the outputs from model
        if run_successful:
            for var_name, specs in output_dict["outputs"].items():
                pyo_obj = model.find_component(specs["full_name"])
                # incase value is not initlized or can't be evalauted
                # typical case, is a var is created, but not initlized or touched, such is 0 index vars in 1D RO
                try:
                    output_dict["outputs"][var_name]["value"][case_number] = pyo.value(
                        pyo_obj
                    )
                except ValueError:
                    pass

        else:
            for label, specs in output_dict["outputs"].items():
                pyo_obj = model.find_component(specs["full_name"])
                if pyo_obj.name in sweep_params.keys():
                    output_dict["outputs"][label]["value"][case_number] = pyo.value(
                        pyo_obj
                    )
                else:
                    output_dict["outputs"][label]["value"][case_number] = np.nan

    def _create_global_output(self, local_output_dict, req_num_samples=None):
        # We make the assumption that the parameter sweep is running the same
        # flowsheet num_samples number of times, i.e., the structure of the
        # local_output_dict remains the same across all mpi_ranks
        local_num_cases = len(local_output_dict["solve_successful"])

        # Gather the size of the value array for each peer process
        sample_split_arr = self.parallel_manager.combine_data_with_peers(
            local_num_cases
        )

        num_total_samples = sum(sample_split_arr)
        if req_num_samples is None:
            req_num_samples = num_total_samples

        # Create the global value array on rank 0
        if self.parallel_manager.is_root_process():
            global_output_dict = copy.deepcopy(local_output_dict)
            # Create a global value array of inputs in the dictionary
            for key, item in global_output_dict.items():
                if key in ["sweep_params", "outputs"]:
                    for subkey, subitem in item.items():
                        subitem["value"] = np.zeros(num_total_samples, dtype=float)

        else:
            global_output_dict = local_output_dict

        # Finally collect the values
        for key, item in local_output_dict.items():
            if key in ["sweep_params", "outputs"]:
                for subkey, subitem in item.items():
                    self.parallel_manager.gather_arrays_to_root(
                        sendbuf=subitem["value"],
                        recvbuf_spec=(
                            global_output_dict[key][subkey]["value"],
                            sample_split_arr,
                        ),
                    )

                    # Trim to the exact number
                    global_output_dict[key][subkey]["value"] = global_output_dict[key][
                        subkey
                    ]["value"][0:req_num_samples]

            elif key == "solve_successful":
                local_solve_successful = np.fromiter(item, dtype=bool, count=len(item))

                if self.parallel_manager.is_root_process():
                    global_solve_successful = np.empty(num_total_samples, dtype=bool)
                else:
                    global_solve_successful = None

                self.parallel_manager.gather_arrays_to_root(
                    sendbuf=local_solve_successful,
                    recvbuf_spec=(global_solve_successful, sample_split_arr),
                )

                if self.parallel_manager.is_root_process():
                    # Trim to the exact number
                    global_output_dict[key] = list(
                        global_solve_successful[0:req_num_samples]
                    )

        return global_output_dict

    def _param_sweep_kernel(self, sweep_params, local_value_k):
        # Forced reinitialization of the flowsheet if enabled
        # and model is not already initalized at givel local sweep param set
        # or init if model was not initialized or prior solved failed (if solved failed, init state is false)
        if (
            self.config.initialize_before_sweep
            and all(self.model_manager.current_k == local_value_k) == False
        ) or self.model_manager.is_initialized == False:
            if self.model_manager._is_rebuild_and_init_enabled:
                self.model_manager.build_and_init(sweep_params, local_value_k)
        # try to solve our model
        self.model_manager.update_model_params(sweep_params, local_value_k)
        self.model_manager.solve_model()

        # if model failed to solve from a prior paramter solved state, lets try
        # to re-init and solve again
        if (
            self.model_manager.is_solved == False
            and self.model_manager.is_prior_parameter_solved == True
            and self.model_manager._is_rebuild_and_init_enabled
        ):
            self.model_manager.build_and_init(sweep_params, local_value_k)
            self.model_manager.update_model_params(sweep_params, local_value_k)
            self.model_manager.solve_model()
        # return model solved state
        return self.model_manager.is_solved

    def _run_sample(
        self,
        local_value_k,
        k,
        sweep_params,
        local_output_dict,
    ):
        # Update model parmeters for record keeping and probe testing
        self._update_model_values(self.model_manager.model, sweep_params, local_value_k)

        if self.config.probe_function is None or self.config.probe_function(
            self.model_manager.model
        ):
            run_successful = self._param_sweep_kernel(
                sweep_params,
                local_value_k,
            )
        else:
            run_successful = False
            # makes sure that if model was build,, but failed to init
            # we store the pars that were run

        # Update the loop based on the reinitialization
        self._update_local_output_dict(
            self.model_manager.model,
            sweep_params,
            k,
            run_successful,
            local_output_dict,
        )
        return run_successful

    def _do_param_sweep(self, sweep_params, outputs, local_values):
        # setup model manager if not already specifid (Used in case of diff tool)
        # or if user wants to specify thier own model_manager before runing param sweep
        if self.model_manager == None:
            self.model_manager = ModelManager(self)

        # build and init model, we also pass first set of paramters incase user wants
        # to update them before initlizeing the model
        if (
            self.config.initialize_before_sweep
            or self.model_manager.is_initialized == False
        ):
            if self.model_manager._is_rebuild_and_init_enabled:
                self.model_manager.build_and_init(
                    sweep_params=sweep_params, local_value_k=local_values[0, :]
                )

        local_num_cases = np.shape(local_values)[0]

        # Create the output skeleton for storing detailed data
        local_output_dict = self._create_local_output_skeleton(
            self.model_manager.model, sweep_params, outputs, local_num_cases
        )

        local_solve_successful_list = []

        # ================================================================
        # Run all optimization cases
        # ================================================================

        for k in range(local_num_cases):
            start_time = time.time()
            run_successful = self._run_sample(
                local_values[k, :],
                k,
                sweep_params,
                local_output_dict,
            )
            time_elapsed = time.time() - start_time
            local_solve_successful_list.append(run_successful)
            self._publish_updates(k, run_successful, time_elapsed)

        local_output_dict["solve_successful"] = local_solve_successful_list

        return local_output_dict

    @abstractmethod
    def parameter_sweep(self, *args, **kwargs):
        pass


class ParameterSweep(_ParameterSweepBase, _ParameterSweepParallelUtils):
    CONFIG = _ParameterSweepBase.CONFIG()

    def parameter_sweep(
        self,
        build_model,
        build_sweep_params,
        build_outputs=None,
        build_outputs_kwargs=None,
        num_samples=None,
        seed=None,
        build_model_kwargs=None,
        build_sweep_params_kwargs=None,
    ):
        build_model_kwargs = (
            build_model_kwargs if build_model_kwargs is not None else dict()
        )
        build_outputs_kwargs = (
            build_outputs_kwargs if build_outputs_kwargs is not None else dict()
        )
        build_sweep_params_kwargs = (
            build_sweep_params_kwargs
            if build_sweep_params_kwargs is not None
            else dict()
        )

        if not callable(build_model):
            _model = build_model
            build_model = lambda: _model
            deprecation_warning(
                "Passing a model directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )

        if not callable(build_sweep_params):
            _sweep_params = build_sweep_params
            build_sweep_params = lambda model: _sweep_params
            deprecation_warning(
                "Passing sweep params directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )

        if build_outputs is None:
            build_outputs = return_none

        if not callable(build_outputs):
            _combined_outputs = build_outputs
            build_outputs = lambda model: _combined_outputs
            deprecation_warning(
                "Passing the output dict directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )
        # This should be depreciated in future versions
        self.config.build_model = build_model
        self.config.build_sweep_params = build_sweep_params
        self.config.build_outputs = build_outputs
        self.config.build_outputs_kwargs = build_outputs_kwargs
        self.config.build_model_kwargs = build_model_kwargs
        self.config.build_sweep_params_kwargs = build_sweep_params_kwargs
        # create the list of all combinations - needed for some aspects of scattering
        model = build_model(**build_model_kwargs)
        sweep_params = build_sweep_params(model, **build_sweep_params_kwargs)
        sweep_params, sampling_type = self._process_sweep_params(sweep_params)
        np.random.seed(seed)
        all_parameter_combinations = self._build_combinations(
            sweep_params, sampling_type, num_samples
        )

        all_results = self.run_scatter_gather(
            all_parameter_combinations, ParameterSweep
        )

        global_sweep_results_dict = self._combine_gather_results(all_results)
        combined_output_arr = self._combine_output_array(global_sweep_results_dict)
        all_parameter_combinations_solved = self._combine_input_array(
            global_sweep_results_dict
        )
        # save the results for all simulations run by this process and its children
        for results in self.parallel_manager.results_from_local_tree(all_results):
            self.writer.save_results(
                sweep_params,
                results.parameters,
                all_parameter_combinations_solved,
                results.results,
                global_sweep_results_dict,
                combined_output_arr,
                process_number=results.process_number,
            )

        global_save_data = np.hstack((all_parameter_combinations, combined_output_arr))

        return global_save_data, global_sweep_results_dict


class RecursiveParameterSweep(_ParameterSweepBase):
    CONFIG = _ParameterSweepBase.CONFIG()

    def _filter_recursive_solves(
        self, model, sweep_params, outputs, recursive_local_dict
    ):
        # Figure out how many filtered solves did this rank actually do
        filter_counter = 0
        for case, content in recursive_local_dict.items():
            filter_counter += sum(content["solve_successful"])

        # Now that we have all of the local output dictionaries, we need to construct
        # a consolidated dictionary of successful solves.
        local_filtered_dict = self._create_local_output_skeleton(
            model, sweep_params, outputs, filter_counter
        )
        local_filtered_dict["solve_successful"] = []

        # Populate local_successful_outputs
        offset = 0
        for case_number, content in recursive_local_dict.items():
            # Filter all of the sucessful solves
            optimal_indices = [
                idx
                for idx, success in enumerate(content["solve_successful"])
                if success
            ]
            n_successful_solves = len(optimal_indices)
            stop = offset + n_successful_solves

            for key, item in content.items():
                if key != "solve_successful":
                    for subkey, subitem in item.items():
                        local_filtered_dict[key][subkey]["value"][
                            offset:stop
                        ] = subitem["value"][optimal_indices]

            # Place the solve status
            local_filtered_dict["solve_successful"].extend(
                [content["solve_successful"][i] for i in optimal_indices]
            )

            offset += n_successful_solves

        return local_filtered_dict, filter_counter

    def _aggregate_filtered_input_arr(self, global_filtered_dict, req_num_samples):
        global_filtered_values = np.zeros(
            (req_num_samples, len(global_filtered_dict["sweep_params"])),
            dtype=float,
        )

        if self.parallel_manager.is_root_process():
            for i, (key, item) in enumerate(
                global_filtered_dict["sweep_params"].items()
            ):
                global_filtered_values[:, i] = item["value"][:req_num_samples]

        self.parallel_manager.sync_array_with_peers(global_filtered_values)

        return global_filtered_values

    def _aggregate_filtered_results(self, local_filtered_dict, req_num_samples):
        global_filtered_dict = self._create_global_output(
            local_filtered_dict, req_num_samples
        )
        global_filtered_results = self._aggregate_results_arr(
            global_filtered_dict, req_num_samples
        )
        global_filtered_values = self._aggregate_filtered_input_arr(
            global_filtered_dict, req_num_samples
        )

        return global_filtered_dict, global_filtered_results, global_filtered_values

    @staticmethod
    def _update_sweep_params(sweep_params, num_total_samples):
        for obj in sweep_params.values():
            obj.num_samples = num_total_samples

    def parameter_sweep(
        self,
        build_model,
        build_sweep_params,
        build_outputs=None,
        build_outputs_kwargs=None,
        num_samples=None,
        seed=None,
        build_model_kwargs=None,
        build_sweep_params_kwargs=None,
    ):
        build_model_kwargs = (
            build_model_kwargs if build_model_kwargs is not None else dict()
        )
        build_outputs_kwargs = (
            build_outputs_kwargs if build_outputs_kwargs is not None else dict()
        )
        build_sweep_params_kwargs = (
            build_sweep_params_kwargs
            if build_sweep_params_kwargs is not None
            else dict()
        )

        if not callable(build_model):
            _model = build_model
            build_model = lambda: _model
            deprecation_warning(
                "Passing a model directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )

        if not callable(build_sweep_params):
            _sweep_params = build_sweep_params
            build_sweep_params = lambda model: _sweep_params
            deprecation_warning(
                "Passing sweep params directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )

        if build_outputs is None:
            build_outputs = return_none

        if not callable(build_outputs):
            _combined_outputs = build_outputs
            build_outputs = lambda model: _combined_outputs
            deprecation_warning(
                "Passing the output dict directly to the parameter_sweep function is deprecated \
                                and will not work with future implementations of parallelism.",
                version="0.10.0",
            )
        # This should be depreciated in future versions
        self.config.build_model = build_model
        self.config.build_sweep_params = build_sweep_params
        self.config.build_outputs = build_outputs
        self.config.build_outputs_kwargs = build_outputs_kwargs
        self.config.build_model_kwargs = build_model_kwargs
        self.config.build_sweep_params_kwargs = build_sweep_params_kwargs
        # create the list of all combinations - needed for some aspects of scattering
        model = build_model(**build_model_kwargs)
        sweep_params = build_sweep_params(model, **build_sweep_params_kwargs)
        sweep_params, sampling_type = self._process_sweep_params(sweep_params)
        outputs = build_outputs(model, **build_outputs_kwargs)
        # Set the seed before sampling
        np.random.seed(seed)

        # Check if the outputs have the name attribute. If not, assign one.
        if outputs is not None:
            self.assign_variable_names(model, outputs)

        n_samples_remaining = num_samples
        num_total_samples = num_samples

        local_output_collection = {}
        for loop_ctr in range(10):
            if n_samples_remaining <= 0:
                break

            if loop_ctr > 0:
                # We need to rebuild the sweep_params since these are single use objects
                self._update_sweep_params(sweep_params, num_total_samples)

            # Enumerate/Sample the parameter space
            global_values = self._build_combinations(
                sweep_params, sampling_type, num_total_samples
            )

            # divide the workload between processors
            local_values = self._divide_combinations(global_values)
            local_num_cases = np.shape(local_values)[0]
            if loop_ctr == 0:
                true_local_num_cases = local_num_cases

            if self.config.custom_do_param_sweep is None:
                local_output_collection[loop_ctr] = self._do_param_sweep(
                    sweep_params,
                    outputs,
                    local_values,
                )
            else:
                local_output_collection[
                    loop_ctr
                ] = self.self.config.custom_do_param_sweep(
                    sweep_params,
                    outputs,
                    local_values,
                    **self.config.custom_do_param_sweep_kwargs,
                )

            # Get the number of successful solves on this proc (sum of boolean flags)
            success_count = sum(local_output_collection[loop_ctr]["solve_successful"])
            failure_count = local_num_cases - success_count

            # Get the global number of successful solves and update the number of remaining samples
            if (
                self.parallel_manager.number_of_worker_processes() > 1
            ):  # pragma: no cover
                global_success_count = np.zeros(1, dtype=int)
                global_failure_count = np.zeros(1, dtype=int)

                self.parallel_manager.sum_values_and_sync(
                    sendbuf=np.array(success_count, dtype=int),
                    recvbuf=global_success_count,
                )

                self.parallel_manager.sum_values_and_sync(
                    sendbuf=np.array(failure_count, dtype=int),
                    recvbuf=global_failure_count,
                )
            else:
                global_success_count = success_count
                global_failure_count = failure_count

            success_prob = global_success_count / (
                global_failure_count + global_success_count
            )

            if success_prob < 0.1:
                warnings.warn(
                    f"Success rate of solves = {100.0*success_prob}%, consider adjusting sweep limits."
                )

            n_samples_remaining -= global_success_count

            # The total number of samples to generate at the next iteration is a multiple of the total remaining samples
            scale_factor = 2.0 / max(success_prob, 0.10)
            num_total_samples = int(np.ceil(scale_factor * n_samples_remaining))

        # Now that we have all of the local output dictionaries, we need to construct
        # a consolidated dictionary based on a filter, e.g., optimal solves.
        local_filtered_dict, local_n_successful = self._filter_recursive_solves(
            model, sweep_params, outputs, local_output_collection
        )

        # if we are debugging
        if self.writer.config["debugging_data_dir"] is not None:
            local_filtered_values = np.zeros(
                (local_n_successful, len(local_filtered_dict["sweep_params"])),
                dtype=float,
            )
            for i, (key, item) in enumerate(
                local_filtered_dict["sweep_params"].items()
            ):
                local_filtered_values[:, i] = item["value"][:]
        else:
            local_filtered_values = None

        # Not that we have all of the successful outputs in a consolidated dictionary locally,
        # we can now construct a global dictionary of successful solves.
        (
            global_filtered_dict,
            global_filtered_results,
            global_filtered_values,
        ) = self._aggregate_filtered_results(local_filtered_dict, num_samples)

        # Now we can save this
        self.parallel_manager.sync_with_peers()

        # Save to file
        global_save_data = self.writer.save_results(
            sweep_params,
            local_filtered_values,
            global_filtered_values,
            local_filtered_dict,
            global_filtered_dict,
            global_filtered_results,
            self.parallel_manager.get_rank(),
        )

        return global_save_data, global_filtered_dict



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    NonNegativeReals,
    exp,
    log10,
    Param,
    Var,
    check_optimal_termination,
    Set,
    Suffix,
    units as pyunits,
    Constraint,
)
from idaes.core.util.math import smooth_max
from pyomo.common.config import ConfigBlock, ConfigValue, In
from enum import Enum, auto

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.misc import add_object_reference

from watertap.core import InitializationMixin
from watertap.costing.unit_models.uv_aop import cost_uv_aop

_log = idaeslog.getLogger(__name__)

# ---------------------------------------------------------------------
class UVDoseType(Enum):
    fixed = auto()  # uv dose is a user specified value
    calculated = auto()


@declare_process_block_class("Ultraviolet0D")
class Ultraviolet0DData(InitializationMixin, UnitModelBlockData):
    """
    Standard UV Unit Model Class:
    - zero dimensional model
    - steady state only
    - single liquid phase only
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. UV units do not support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. UV units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "uv_dose_type",
        ConfigValue(
            default=UVDoseType.fixed,
            domain=In(UVDoseType),
            description="UV dose construction flag",
            doc="""Indicates whether the uv dose will be calculated or fixed by the user
        **default** - UVDoseType.fixed.
        **Valid values:** {
        **UVDoseType.fixed** - user specifies uv dose,
        **UVDoseType.calculated** - calculates uv dose based on the lamp power and UV transmittance}""",
        ),
    )
    CONFIG.declare(
        "has_aop",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="AOP term construction flag",
            doc="""Indicates whether terms for AOP should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include AOP terms,
    **False** - exclude AOP terms.}""",
        ),
    )
    CONFIG.declare(
        "target_species",
        ConfigValue(
            default=None,
            domain=set,
            description="Species target for uv disinfection",
            doc="""Indicate which component in the property package's component list is the target species
        for disinfection by the UV system, currently the model supports a single species
        **default** - None.
        **Valid values:** {
        if the property package solute set only contains one item (two component, one solute, one solvent/water),
        the model will accept the single solute as the target species, for multi-solute systems a string of
        the component id must be provided.}""",
        ),
    )

    def build(self):
        # Call UnitModel.build to setup dynamics
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        if list(self.config.property_package.phase_list) != ["Liq"]:
            raise ConfigurationError(
                "UV model only supports one liquid phase ['Liq'],"
                "the property package has specified the following phases {}".format(
                    [p for p in self.config.property_package.phase_list]
                )
            )

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # separate target species to be adsorbed and other species considered inert
        component_set = self.config.property_package.component_list
        solute_set = self.config.property_package.solute_set
        # apply target species automatically if arg left to default and only one viable option exists
        if self.config.target_species is None and len(solute_set) == 1:
            add_object_reference(self, "target_species", solute_set)
        elif self.config.target_species is not None:
            self.target_species = Set()
            for k in self.config.target_species:
                self.target_species.add(k)
            self.inert_species = component_set - self.target_species
        else:
            raise ConfigurationError(
                "Target species should be a set of strings. Please provide target species and a correct format"
            )

        # Add unit parameters
        self.inactivation_rate = Var(
            self.config.property_package.phase_list,
            self.target_species,
            initialize=2.5e-4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time") ** 2 * units_meta("mass") ** -1,
            doc="Inactivation rate coefficient with respect to uv dose.",
        )

        self.rate_constant = Var(
            self.config.property_package.phase_list,
            self.target_species,
            initialize=2.5e-3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time") ** -1,
            doc="Overall pseudo-first order rate constant.",
        )

        self.photolysis_rate_constant = Var(
            self.config.property_package.phase_list,
            self.target_species,
            initialize=2e-3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time") ** -1,
            doc="Pseudo-first-order rate constant for direct photolysis of component.",
        )

        self.reaction_rate_constant = Var(
            self.config.property_package.phase_list,
            self.target_species,
            initialize=5e-4,
            bounds=(0, 100),
            domain=NonNegativeReals,
            units=units_meta("time") ** -1,
            doc="Pseudo-first-order rate constant for indirect photolysis of component.",
        )

        self.dens_solvent = Param(
            initialize=1000,
            units=units_meta("mass") * units_meta("length") ** -3,
            doc="Pure water density",
        )

        # Add uv variables
        self.uv_dose = Var(
            initialize=5000,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -2,
            doc="UV dose.",
        )
        self.uv_intensity = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -3,
            doc="Average intensity of UV light.",
        )
        self.exposure_time = Var(
            initialize=500,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="Exposure time of UV light.",
        )
        self.reactor_volume = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 3,
            doc="UV reactor volume.",
        )

        # Add electricity parameters
        self.electricity_demand_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.target_species,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Electricity demand per component",
        )

        self.max_phase_electricity_demand = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.target_species,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Maximum electricity demand for components with different phases",
        )

        self.electricity_demand_comp = Var(
            self.flowsheet().config.time,
            self.target_species,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Electricity demand per component",
        )

        self.max_component_electricity_demand = Var(
            self.flowsheet().config.time,
            self.target_species,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Maximum electricity demand for multiple components",
        )

        self.electricity_demand = Var(
            self.flowsheet().config.time,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Electricity demand of unit",
        )

        self.eps_electricity = Param(
            mutable=True,
            initialize=1e-3,
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("length") ** 2
            * units_meta("time") ** -3,
            doc="Smoothing term for maximum electricity demand",
        )

        self.electrical_efficiency_phase_comp = Var(
            self.flowsheet().time,
            self.config.property_package.phase_list,
            self.target_species,
            initialize=1,
            bounds=(0, None),
            units=units_meta("mass")
            * units_meta("length") ** -1
            * units_meta("time") ** -2,
            doc="Electricity efficiency per log order reduction (EE/O)",
        )

        self.lamp_efficiency = Var(
            initialize=0.3,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Lamp efficiency",
        )

        # Build control volume for UV unit
        self.control_volume = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_mass_transfer=True
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_enthalpy_transfer=True
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Add Ports
        self.add_inlet_port(name="inlet", block=self.control_volume)
        self.add_outlet_port(name="outlet", block=self.control_volume)

        # References for control volume
        # pressure change
        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # UV dose
        @self.Constraint(
            doc="Constraint for UV dose",
        )
        def eq_uv_dose(b):
            return b.uv_dose == b.uv_intensity * b.exposure_time

        # ---------------------------------------------------------------------
        if self.config.uv_dose_type == UVDoseType.calculated:
            self.UVT = Var(
                initialize=0.9,
                bounds=(0, 1),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="UV transmittance.",
            )
            self.A_coeff = Var(
                initialize=2.15138,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="UV dose delivery model coefficient A.",
            )
            self.B_coeff = Var(
                initialize=10.2072,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="UV dose delivery model coefficient B.",
            )
            self.C_coeff = Var(
                initialize=0.696709,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="UV dose delivery model coefficient C.",
            )
            self.D_coeff = Var(
                initialize=1.09563,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="UV dose delivery model coefficient D.",
            )
            self.relative_lamp_output = Var(
                initialize=1,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Output from the lamps relative to a new lamp, S/S0.",
            )
            self.num_of_banks = Var(
                initialize=2,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Number of banks.",
            )

            @self.Constraint(
                doc="Constraint for UV dose based on lamp power and UV transmittance",
            )
            def eq_uv_dose_detail(b):
                Qr = pyunits.convert(
                    pyunits.convert(
                        b.control_volume.properties_in[0].flow_vol,
                        to_units=pyunits.gal / pyunits.day,
                    )
                    / (1000000 * pyunits.gal / pyunits.day),
                    to_units=pyunits.dimensionless,
                )
                UVA = -log10(b.UVT)
                return b.uv_dose == pyunits.convert(
                    (
                        10**b.A_coeff
                        * UVA ** (b.B_coeff * UVA)
                        * (b.relative_lamp_output / Qr) ** b.C_coeff
                        * b.num_of_banks**b.D_coeff
                        * pyunits.mJ
                        / pyunits.cm**2
                    ),
                    to_units=units_meta("mass") * units_meta("time") ** -2,
                )

        # UV reactor volume
        @self.Constraint(
            doc="Constraint for UV reactor volume",
        )
        def eq_uv_reactor_volume(b):
            t0 = b.flowsheet().time.first()
            return (
                b.reactor_volume
                == b.control_volume.properties_in[t0].flow_vol * b.exposure_time
            )

        # rate constant
        @self.Constraint(
            self.config.property_package.phase_list,
            self.target_species,
            doc="Constraint for pseudo-first order rate constant with respect to uv intensity",
        )
        def eq_rate_constant(b, p, j):
            return b.rate_constant[p, j] == b.uv_intensity * b.inactivation_rate[p, j]

        @self.Constraint(
            self.config.property_package.phase_list,
            self.target_species,
            doc="Constraint for pseudo-first order reaction rate constant with respect to direct and indirect photolysis",
        )
        def eq_overall_rate_constant(b, p, j):
            return (
                b.rate_constant[p, j]
                == b.photolysis_rate_constant[p, j] + b.reaction_rate_constant[p, j]
            )

        # ---------------------------------------------------------------------
        if self.config.has_aop is True:
            self.second_order_reaction_rate_constant = Var(
                self.config.property_package.phase_list,
                self.target_species,
                initialize=3.3e8,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.M**-1 * pyunits.s**-1,
                doc="Second-order reaction rate constant for the reaction between component and hydrogen peroxide.",
            )
            self.hydrogen_peroxide_conc = Var(
                initialize=5.05e-13,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.M,
                doc="Steady-state concentration of hydrogen peroxide.",
            )

            @self.Constraint(
                self.config.property_package.phase_list,
                self.target_species,
                doc="Constraint for pseudo-first order reaction rate constant",
            )
            def eq_reaction_rate_constant(b, p, j):
                return (
                    b.reaction_rate_constant[p, j]
                    == b.second_order_reaction_rate_constant[p, j]
                    * b.hydrogen_peroxide_conc
                )

        else:
            self.reaction_rate_constant.fix(0)

        # mass transfer
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Constraints for solvent and solute concentration in outlet stream.",
        )
        def eq_outlet_conc(b, t, p, j):
            prop_in = b.control_volume.properties_in[t]
            if j in b.target_species:
                return b.control_volume.mass_transfer_term[t, p, j] == -(
                    prop_in.get_material_flow_terms(p, j)
                    * (
                        1
                        - exp(
                            pyunits.convert(
                                -b.uv_dose * b.inactivation_rate[p, j],
                                to_units=pyunits.dimensionless,
                            )
                        )
                    )
                )
            else:
                b.control_volume.mass_transfer_term[t, p, j].fix(0)
                return Constraint.Skip

        # electricity
        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.target_species,
            doc="Constraints for electricity demand of the UV reactor.",
        )
        def eq_electricity_demand_phase_comp(b, t, p, j):
            prop_in = b.control_volume.properties_in[t]
            return b.electricity_demand_phase_comp[t, p, j] == (
                b.electrical_efficiency_phase_comp[t, p, j]
                * prop_in.flow_vol
                * log10(
                    1
                    / exp(
                        pyunits.convert(
                            -b.uv_dose * b.inactivation_rate[p, j],
                            to_units=pyunits.dimensionless,
                        )
                    )
                )
                / b.lamp_efficiency
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.target_species,
            doc="Constraints for calculating maximum electricity demand for different phases.",
        )
        def eq_max_phase_electricity_demand(b, t, p, j):
            if p == b.config.property_package.phase_list.first():
                return (
                    b.max_phase_electricity_demand[t, p, j]
                    == b.electricity_demand_phase_comp[t, p, j]
                )
            else:
                return b.max_phase_electricity_demand[t, p, j] == (
                    smooth_max(
                        b.max_phase_electricity_demand[
                            t, b.config.property_package.phase_list.prev(p), j
                        ],
                        b.electricity_demand_phase_comp[t, p, j],
                        b.eps_electricity,
                    )
                )

        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="Constraints for electricity demand of each component.",
        )
        def eq_electricity_demand_comp(b, t, j):
            return (
                b.electricity_demand_comp[t, j]
                == b.max_phase_electricity_demand[
                    t, b.config.property_package.phase_list.last(), j
                ]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="Constraints for calculating maximum electricity demand for multiple components.",
        )
        def eq_max_electricity_demand(b, t, j):
            if j == b.target_species.first():
                return (
                    b.max_component_electricity_demand[t, j]
                    == b.electricity_demand_comp[t, j]
                )
            else:
                return b.max_component_electricity_demand[t, j] == (
                    smooth_max(
                        b.max_component_electricity_demand[t, b.target_species.prev(j)],
                        b.electricity_demand_comp[t, j],
                        b.eps_electricity,
                    )
                )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Constraints for total electricity demand of the UV reactor.",
        )
        def eq_electricity_demand(b, t):
            return (
                b.electricity_demand[t]
                == b.max_component_electricity_demand[t, b.target_species.last()]
            )

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags = self.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Initialize permeate
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = self.control_volume.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        init_log.info_high("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        self.control_volume.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_performance_contents(self, time_point=0):
        # TODO: add other performance constants
        var_dict = {}

        var_dict["UV dose"] = self.uv_dose
        var_dict["Average intensity of UV light"] = self.uv_intensity

        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        # loop through desired state block properties indexed by [phase, comp]
        phase_comp_prop_dict = {
            "flow_mol_phase_comp": "Molar flow rate",
            "flow_mass_phase_comp": "Mass flow rate",
            "conc_mol_phase_comp": "Molar concentration",
            "conc_mass_phase_comp": "Mass concentration",
        }
        for prop_name, prop_str in phase_comp_prop_dict.items():
            for j in self.config.property_package.component_list:
                if self.control_volume.properties_in[
                    time_point
                ].is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ process feed inlet"] = getattr(
                        self.control_volume.properties_in[time_point], prop_name
                    )["Liq", j]
                if self.control_volume.properties_out[
                    time_point
                ].is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ process feed outlet"] = getattr(
                        self.control_volume.properties_out[time_point], prop_name
                    )["Liq", j]

        # loop through desired state block properties indexed by [phase]
        phase_prop_dict = {
            "flow_vol_phase": "Volumetric flow rate",
        }
        for prop_name, prop_str in phase_prop_dict.items():
            if self.control_volume.properties_in[time_point].is_property_constructed(
                prop_name
            ):
                var_dict[f"{prop_str} @ process feed inlet"] = getattr(
                    self.control_volume.properties_in[time_point], prop_name
                )["Liq"]
            if self.control_volume.properties_out[time_point].is_property_constructed(
                prop_name
            ):
                var_dict[f"{prop_str} @ process feed outlet"] = getattr(
                    self.control_volume.properties_out[time_point], prop_name
                )["Liq"]

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Process Inlet": self.inlet,
                "Process Outlet": self.outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables
        # these variables should have user input, if not there will be a warning
        if iscale.get_scaling_factor(self.uv_intensity) is None:
            sf = iscale.get_scaling_factor(self.uv_intensity, default=0.1, warning=True)
        iscale.set_scaling_factor(self.uv_intensity, sf)

        if iscale.get_scaling_factor(self.exposure_time) is None:
            sf = iscale.get_scaling_factor(
                self.exposure_time, default=1e-2, warning=True
            )
            iscale.set_scaling_factor(self.exposure_time, sf)

        if iscale.get_scaling_factor(self.uv_dose) is None:
            sf = iscale.get_scaling_factor(self.uv_dose, default=1e-3, warning=True)
        iscale.set_scaling_factor(self.uv_dose, sf)

        if iscale.get_scaling_factor(self.inactivation_rate) is None:
            sf = iscale.get_scaling_factor(
                self.inactivation_rate, default=1e4, warning=True
            )
        iscale.set_scaling_factor(self.inactivation_rate, sf)

        if iscale.get_scaling_factor(self.rate_constant) is None:
            sf = iscale.get_scaling_factor(
                self.rate_constant, default=1e3, warning=True
            )
        iscale.set_scaling_factor(self.rate_constant, sf)

        if iscale.get_scaling_factor(self.electrical_efficiency_phase_comp) is None:
            sf = iscale.get_scaling_factor(
                self.electrical_efficiency_phase_comp, default=1e-5, warning=True
            )
        iscale.set_scaling_factor(self.electrical_efficiency_phase_comp, sf)

        if hasattr(self, "photolysis_rate_constant"):
            if iscale.get_scaling_factor(self.photolysis_rate_constant) is None:
                sf = iscale.get_scaling_factor(
                    self.photolysis_rate_constant, default=1e3, warning=True
                )
            iscale.set_scaling_factor(self.photolysis_rate_constant, sf)

        if hasattr(self, "reaction_rate_constant"):
            if iscale.get_scaling_factor(self.reaction_rate_constant) is None:
                sf = iscale.get_scaling_factor(
                    self.reaction_rate_constant, default=1e3, warning=True
                )
            iscale.set_scaling_factor(self.reaction_rate_constant, sf)

        if hasattr(self, "second_order_reaction_rate_constant"):
            if (
                iscale.get_scaling_factor(self.second_order_reaction_rate_constant)
                is None
            ):
                sf = iscale.get_scaling_factor(
                    self.second_order_reaction_rate_constant, default=1e-8, warning=True
                )
            iscale.set_scaling_factor(self.second_order_reaction_rate_constant, sf)

        if hasattr(self, "hydrogen_peroxide_conc"):
            if iscale.get_scaling_factor(self.hydrogen_peroxide_conc) is None:
                sf = iscale.get_scaling_factor(
                    self.hydrogen_peroxide_conc, default=1e13, warning=True
                )
            iscale.set_scaling_factor(self.hydrogen_peroxide_conc, sf)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor
        if iscale.get_scaling_factor(self.lamp_efficiency) is None:
            iscale.set_scaling_factor(self.lamp_efficiency, 1)

        if hasattr(self, "UVT"):
            if iscale.get_scaling_factor(self.UVT) is None:
                iscale.set_scaling_factor(self.UVT, 1)

        if hasattr(self, "A_coeff"):
            if iscale.get_scaling_factor(self.A_coeff) is None:
                iscale.set_scaling_factor(self.A_coeff, 1)

        if hasattr(self, "B_coeff"):
            if iscale.get_scaling_factor(self.B_coeff) is None:
                iscale.set_scaling_factor(self.B_coeff, 1)

        if hasattr(self, "C_coeff"):
            if iscale.get_scaling_factor(self.C_coeff) is None:
                iscale.set_scaling_factor(self.C_coeff, 1)

        if hasattr(self, "D_coeff"):
            if iscale.get_scaling_factor(self.D_coeff) is None:
                iscale.set_scaling_factor(self.D_coeff, 1)

        if hasattr(self, "relative_lamp_output"):
            if iscale.get_scaling_factor(self.relative_lamp_output) is None:
                iscale.set_scaling_factor(self.relative_lamp_output, 1)

        if hasattr(self, "num_of_banks"):
            if iscale.get_scaling_factor(self.num_of_banks) is None:
                iscale.set_scaling_factor(self.num_of_banks, 1)

        if iscale.get_scaling_factor(self.reactor_volume) is None:
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[0].flow_vol
            ) * iscale.get_scaling_factor(self.exposure_time)
            iscale.set_scaling_factor(self.reactor_volume, sf)

        if iscale.get_scaling_factor(self.dens_solvent) is None:
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[0].dens_mass_phase["Liq"]
            )
            iscale.set_scaling_factor(self.dens_solvent, sf)

        if iscale.get_scaling_factor(self.eps_electricity) is None:
            iscale.set_scaling_factor(self.eps_electricity, 1e3)

        for (t, p, j), v in self.electricity_demand_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                removal = -iscale.get_scaling_factor(
                    self.uv_dose
                ) * iscale.get_scaling_factor(self.inactivation_rate[p, j])
                sf = (
                    iscale.get_scaling_factor(
                        self.electrical_efficiency_phase_comp[t, p, j]
                    )
                    * (1 / log10(1 / exp(removal)))
                    * iscale.get_scaling_factor(
                        self.control_volume.properties_in[t].flow_vol
                    )
                    / iscale.get_scaling_factor(self.lamp_efficiency)
                )
                iscale.set_scaling_factor(v, sf)

        for (t, p, j), v in self.max_phase_electricity_demand.items():
            if iscale.get_scaling_factor(v) is None:
                p = self.config.property_package.phase_list.first()
                sf = iscale.get_scaling_factor(
                    self.electricity_demand_phase_comp[t, p, j], warning=True
                )
                iscale.set_scaling_factor(v, sf)

        for (t, j), v in self.electricity_demand_comp.items():
            if iscale.get_scaling_factor(v) is None:
                p = self.config.property_package.phase_list.last()
                sf = iscale.get_scaling_factor(
                    self.max_phase_electricity_demand[t, p, j], warning=True
                )
                iscale.set_scaling_factor(v, sf)

        for (t, j), v in self.max_component_electricity_demand.items():
            if iscale.get_scaling_factor(v) is None:
                j = self.target_species.first()
                sf = iscale.get_scaling_factor(
                    self.electricity_demand_comp[t, j], warning=True
                )
                iscale.set_scaling_factor(v, sf)

        for t, v in self.electricity_demand.items():
            if iscale.get_scaling_factor(v) is None:
                j = self.target_species.last()
                sf = iscale.get_scaling_factor(
                    self.max_component_electricity_demand[t, j], warning=True
                )
                iscale.set_scaling_factor(v, sf)

        # TODO: update IDAES control volume to scale mass_transfer and enthalpy_transfer
        for ind, v in self.control_volume.mass_transfer_term.items():
            (t, p, j) = ind
            if iscale.get_scaling_factor(v) is None:
                sf = iscale.get_scaling_factor(
                    self.control_volume.mass_transfer_term[t, p, j]
                )
                iscale.constraint_scaling_transform(
                    self.control_volume.material_balances[t, j], sf
                )

        # transforming constraints
        for c in self.eq_uv_dose.values():
            if iscale.get_scaling_factor(self.uv_dose) is None:
                sf = iscale.get_scaling_factor(
                    self.uv_intensity
                ) * iscale.get_scaling_factor(self.exposure_time)
            else:
                sf = iscale.get_scaling_factor(self.uv_dose)
            iscale.constraint_scaling_transform(c, sf)

        for c in self.eq_uv_reactor_volume.values():
            sf = iscale.get_scaling_factor(self.reactor_volume)
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_rate_constant.items():
            if iscale.get_scaling_factor(self.rate_constant) is None:
                sf = iscale.get_scaling_factor(
                    self.uv_intensity
                ) * iscale.get_scaling_factor(self.inactivation_rate[ind])
            else:
                sf = iscale.get_scaling_factor(self.rate_constant[ind])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_overall_rate_constant.items():
            if iscale.get_scaling_factor(self.rate_constant) is None:
                sf = iscale.get_scaling_factor(
                    self.photolysis_rate_constant[ind]
                ) + iscale.get_scaling_factor(self.reaction_rate_constant[ind])
            else:
                sf = iscale.get_scaling_factor(self.rate_constant[ind])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_outlet_conc.items():
            (t, p, j) = ind
            sf = iscale.get_scaling_factor(
                self.control_volume.mass_transfer_term[t, p, j]
            )
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_electricity_demand_phase_comp.items():
            (t, p, j) = ind
            sf = iscale.get_scaling_factor(self.electricity_demand_phase_comp[t, p, j])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_max_phase_electricity_demand.items():
            (t, p, j) = ind
            sf = iscale.get_scaling_factor(self.max_phase_electricity_demand[t, p, j])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_electricity_demand_comp.items():
            (t, j) = ind
            sf = iscale.get_scaling_factor(self.electricity_demand_comp[t, j])
            iscale.constraint_scaling_transform(c, sf)

        for ind, c in self.eq_max_electricity_demand.items():
            (t, j) = ind
            sf = iscale.get_scaling_factor(self.max_component_electricity_demand[t, j])
            iscale.constraint_scaling_transform(c, sf)

        for t, c in self.eq_electricity_demand.items():
            sf = iscale.get_scaling_factor(self.electricity_demand[t])
            iscale.constraint_scaling_transform(c, sf)

        if hasattr(self, "eq_uv_dose_detail"):
            for c in self.eq_uv_dose_detail.values():
                sf = iscale.get_scaling_factor(self.uv_dose)
                iscale.constraint_scaling_transform(c, sf)

        if hasattr(self, "eq_reaction_rate_constant"):
            for c in self.eq_reaction_rate_constant.values():
                sf = iscale.get_scaling_factor(self.reaction_rate_constant)
                iscale.constraint_scaling_transform(c, sf)

    @property
    def default_costing_method(self):
        return cost_uv_aop



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
mutable parameters for optimization:
    m.fs.system_recovery_target
    m.fs.max_allowable_pressure
"""
from pyomo.environ import ConcreteModel, TransformationFactory

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.scaling import calculate_scaling_factors

from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)

import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_softening as flowsheet_softening
import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_two_stage as flowsheet_two_stage


def build(m, **kwargs):
    """
    Build a flowsheet with softening pretreatment and RO.
    """

    assert kwargs["RO_base"] == "TDS"
    assert not kwargs["has_desal_feed"]
    flowsheet_softening.build_components(m)
    flowsheet_two_stage.build_components(m, pretrt_type="softening", **kwargs)

    return m


def scale(m, **kwargs):
    flowsheet_softening.scale(m)
    flowsheet_two_stage.scale(m, **kwargs)


def initialize(m, **kwargs):
    flowsheet_softening.initialize(m)
    flowsheet_two_stage.initialize(m, **kwargs)


def report(m, **kwargs):
    flowsheet_softening.report(m)
    flowsheet_two_stage.report(m, **kwargs)


def set_optimization_components(m, system_recovery, **kwargs):
    # unfix variables
    m.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0].unfix()
    m.fs.stoich_softening_mixer_unit.lime_stream.flow_mol.setlb(1e-6 / 74.09e-3)
    m.fs.stoich_softening_mixer_unit.lime_stream.flow_mol.setub(1 / 74.09e-3)

    m.fs.stoich_softening_separator_unit.hardness.setlb(10)
    m.fs.stoich_softening_separator_unit.hardness.setub(10000)

    flowsheet_two_stage.set_optimization_components(m, system_recovery, **kwargs)


def set_up_optimization(m, system_recovery=0.50, **kwargs):
    set_optimization_components(m, system_recovery, **kwargs)
    calculate_scaling_factors(m)
    check_dof(m, 7)


def optimize(m, check_termination=True):
    s = get_solver()
    s.options["bound_relax_factor"] = 1e-20
    s.options["bound_push"] = 1e-20
    return solve_block(m, solver=s, tee=True, fail_flag=check_termination)


def solve_flowsheet(**desal_kwargs):
    if desal_kwargs == {}:
        desal_kwargs = flowsheet_two_stage.desal_kwargs

    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    build(m, **desal_kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    scale(m, **desal_kwargs)
    calculate_scaling_factors(m)

    # initialize
    initialize(m, **desal_kwargs)

    check_dof(m)
    s = get_solver()
    s.options["bound_relax_factor"] = 1e-20
    solve_block(m, solver=s, tee=True, fail_flag=True)

    # report
    print("===================================" "\n          Simulation          ")
    report(m, **desal_kwargs)

    return m


def optimize_flowsheet(system_recovery=0.50, **kwargs):
    m = solve_flowsheet(**kwargs)
    set_up_optimization(m, system_recovery=system_recovery, **kwargs)
    optimize(m)

    print("===================================" "\n       Optimization            ")
    report(m, **flowsheet_two_stage.desal_kwargs)

    return m


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        m = solve_flowsheet()
    else:
        desal_kwargs = flowsheet_two_stage.desal_kwargs
        m = optimize_flowsheet(system_recovery=float(sys.argv[1]), **desal_kwargs)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

__author__ = "Paul Vecchiarelli, Ben Knueven"

from collections import UserDict


class FixedKeysDict(UserDict):
    def __init__(self, d):
        self.data = d

    def __setitem__(self, k, v):
        if k not in self.data:
            raise RuntimeError(f" Key {k} not in dictionary.")
        else:
            self.data[k] = v

    def __delitem__(self, k):
        raise Exception(" Deleting keys not supported for this object.")

    def _check_value(self, k, valid_values):
        if k not in self.data:
            raise RuntimeError(f" Key {k} not in dictionary.")
        else:
            if self.data[k] not in valid_values:
                raise RuntimeError(
                    f" Value {self.data[k]} not a valid value for key {k}."
                )

    def pprint(self):
        print("-------------------------")
        for key, value in self.data.items():
            print(f" {key}\n - {value}\n")


default_oli_input_dict = FixedKeysDict(
    {
        "temperature_unit": "K",  # units
        "pressure_unit": "Pa",
        "concentration_unit": "mg/L",
        "gas_fraction_unit": "mole %",
        "alkalinity_unit": "mg HCO3/L",
        "tic_unit": "mol C/L",
        # essential input parameters for water analysis
        "electroneutrality_value": "DominantIon",
        "reconciliation_value": "EquilCalcOnly",
        "AllowSolidsToForm": False,
        "CalcAlkalnity": False,
        # optional inputs
        # required for certain specifications
        "MakeupIonBaseTag": None,
        "CO2GasFraction": None,
        "pH": None,
        "PhAcidTitrant": "HCL",
        "PhBaseTitrant": "NAOH",
        "Alkalinity": None,
        "AlkalinityPhTitrant": None,
        "AlkalinityTitrationEndPointPh": None,
        "TIC": None,
    }
)

# Full list of available optional inputs available: https://devdocs.olisystems.com/optional-inputs
default_oli_optional_properties = FixedKeysDict(
    {
        "scalingIndex": False,
        "prescalingTendencies": False,
        "prescalingTendenciesRigorous": False,
        "scalingTendencies": False,
        "MBGComposition": True,
        "materialBalanceGroup": True,
    }
)

default_oli_unit_set_info = FixedKeysDict(
    {
        "enthalpy": "J",
        "mass": "kg",
        "pt": "Pa",
        "total": "mg",
        "liq1_phs_comp": "mg",
        "solid_phs_comp": "mg",
        "vapor_phs_comp": "mg",
        "liq2_phs_comp": "mg",
        "combined_phs_comp": "mg",
        "molecularConcentration": "mg/L",
    }
)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Translator block representing the ASM2d/ADM1 interface.

Assumptions:
     * Steady-state only

Model formulated from:

Flores-Alsina, X., Solon, K., Mbamba, C.K., Tait, S., Gernaey, K.V., Jeppsson, U. and Batstone, D.J., 2016.
Modelling phosphorus (P), sulfur (S) and iron (Fe) interactions for dynamic simulations of anaerobic digestion processes.
Water Research, 95, pp.370-382. https://github.com/wwtmodels/Plant-Wide-Models
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.models.unit_models.translator import TranslatorData
from idaes.core.util.config import (
    is_reaction_parameter_block,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

from pyomo.environ import (
    Param,
    units as pyunits,
    check_optimal_termination,
    Set,
    Expr_if,
)

from idaes.core.util.exceptions import InitializationError

from watertap.property_models.activated_sludge.modified_asm2d_reactions import (
    DecaySwitch,
)

__author__ = "Marcus Holly"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Translator_ASM2d_ADM1")
class TranslatorDataASM2dADM1(TranslatorData):
    """
    Translator block representing the ASM2d/ADM1 interface
    """

    CONFIG = TranslatorData.CONFIG()
    CONFIG.declare(
        "inlet_reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "inlet_reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
    **default** - None.
    **Valid values:** {
    **None** - no reaction package,
    **ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "outlet_reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see reaction package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TranslatorDataASM2dADM1, self).build()

        eps = 0
        mw_p = 31 * pyunits.kg / pyunits.kmol
        mw_n = 14 * pyunits.kg / pyunits.kmol
        mw_c = 12 * pyunits.kg / pyunits.kmol

        self.f_sI_xc = Param(
            initialize=eps,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Soluble inerts from composites",
        )
        self.f_xI_xc = Param(
            initialize=0.1,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Particulate inerts from composites",
        )
        self.f_ch_xc = Param(
            initialize=0.275,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Carbohydrates from composites",
        )
        self.f_pr_xc = Param(
            initialize=0.275,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Proteins from composites",
        )
        self.f_li_xc = Param(
            initialize=0.35,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Lipids from composites",
        )
        self.f_XPHA_Sva = Param(
            initialize=0.1,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Valerate from polyhydroxyalkanoates",
        )

        self.f_XPHA_Sbu = Param(
            initialize=0.1,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Butyrate from polyhydroxyalkanoates",
        )

        self.f_XPHA_Spro = Param(
            initialize=0.4,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Propionate from polyhydroxyalkanoates",
        )

        self.f_XPHA_Sac = Param(
            initialize=0.4,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Acetate from polyhydroxyalkanoates",
        )

        self.C_PHA = Param(
            initialize=0.3 / 12,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Carbon content of polyhydroxyalkanoates",
        )
        self.i_PSI = Param(
            initialize=0.00649,
            units=pyunits.dimensionless,
            mutable=True,
            doc="P content of inert soluble COD S_I, [kg P/kg COD]",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality volumetric flow equation",
        )
        def eq_flow_vol_rule(blk, t):
            return blk.properties_out[t].flow_vol == blk.properties_in[t].flow_vol

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality temperature equation",
        )
        def eq_temperature_rule(blk, t):
            return blk.properties_out[t].temperature == blk.properties_in[t].temperature

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality pressure equation",
        )
        def eq_pressure_rule(blk, t):
            return blk.properties_out[t].pressure == blk.properties_in[t].pressure

        # -------------------------------------------Step 1----------------------------------------------------------------#
        @self.Expression(
            self.flowsheet().time,
            doc="Concentration of COD demanding compounds in S_O2",
        )
        def COD_SO2(blk, t):
            return blk.properties_in[t].conc_mass_comp["S_O2"] / (
                (1 - blk.config.inlet_reaction_package.Y_H)
                / blk.config.inlet_reaction_package.Y_H
            )

        @self.Expression(self.flowsheet().time, doc="S_O2 concentration at step 1")
        def SO2_AS1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_O2"]
                - (1 - blk.config.inlet_reaction_package.Y_H)
                / blk.config.inlet_reaction_package.Y_H
                * self.COD_SO2[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_A concentration at step 1")
        def SA_AS1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_A"]
                - (1 / blk.config.inlet_reaction_package.Y_H) * self.COD_SO2[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 1")
        def SNH4_AS1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_NH4"]
                - blk.config.inlet_reaction_package.i_NBM * self.COD_SO2[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 1")
        def SPO4_AS1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_PO4"]
                - blk.config.inlet_reaction_package.i_PBM * self.COD_SO2[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 1")
        def SIC_AS1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_IC"]
                - (-1 / blk.config.inlet_reaction_package.Y_H)
                * self.COD_SO2[t]
                * blk.config.inlet_reaction_package.i_CSA
                + self.COD_SO2[t] * blk.config.inlet_reaction_package.i_CXB
            )

        @self.Expression(self.flowsheet().time, doc="X_H concentration at step 1")
        def XH_AS1(blk, t):
            return blk.properties_in[t].conc_mass_comp["X_H"] + self.COD_SO2[t]

        # -------------------------------------------Step 2----------------------------------------------------------------#
        @self.Expression(
            self.flowsheet().time,
            doc="Concentration of COD demanding compounds in S_NO3",
        )
        def COD_SNO3(blk, t):
            return blk.properties_in[t].conc_mass_comp["S_NO3"] / (
                (1 - blk.config.inlet_reaction_package.Y_H)
                / (
                    blk.config.inlet_reaction_package.i_NOx_N2
                    * blk.config.inlet_reaction_package.Y_H
                )
            )

        @self.Expression(self.flowsheet().time, doc="S_A concentration at step 2")
        def SA_AS2(blk, t):
            return (
                blk.SA_AS1[t]
                - (1 / blk.config.inlet_reaction_package.Y_H) * self.COD_SNO3[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 2")
        def SNH4_AS2(blk, t):
            return (
                blk.SNH4_AS1[t]
                - self.COD_SNO3[t] * blk.config.inlet_reaction_package.i_NBM
            )

        @self.Expression(self.flowsheet().time, doc="S_N2 concentration at step 2")
        def SN2_AS2(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_N2"]
                + (
                    (1 - blk.config.inlet_reaction_package.Y_H)
                    / (
                        blk.config.inlet_reaction_package.i_NOx_N2
                        * blk.config.inlet_reaction_package.Y_H
                    )
                )
                * self.COD_SNO3[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_NO3 concentration at step 2")
        def SNO3_AS2(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_NO3"]
                - (
                    (1 - blk.config.inlet_reaction_package.Y_H)
                    / (
                        blk.config.inlet_reaction_package.i_NOx_N2
                        * blk.config.inlet_reaction_package.Y_H
                    )
                )
                * self.COD_SNO3[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 2")
        def SPO4_AS2(blk, t):
            return (
                blk.SPO4_AS1[t]
                - self.COD_SNO3[t] * blk.config.inlet_reaction_package.i_PBM
            )

        @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 2")
        def SIC_AS2(blk, t):
            return (
                blk.SIC_AS1[t]
                - (-1 / blk.config.inlet_reaction_package.Y_H)
                * self.COD_SNO3[t]
                * blk.config.inlet_reaction_package.i_CSA
                + self.COD_SNO3[t] * blk.config.inlet_reaction_package.i_CXB
            )

        @self.Expression(self.flowsheet().time, doc="X_H concentration at step 2")
        def XH_AS2(blk, t):
            return blk.XH_AS1[t] + self.COD_SNO3[t]

        # -------------------------------------------Step 3----------------------------------------------------------------#
        @self.Expression(
            self.flowsheet().time, doc="Nitrogen demand for soluble inerts"
        )
        def S_ND(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_F"]
                * blk.config.inlet_reaction_package.i_NSF
            )

        @self.Expression(
            self.flowsheet().time, doc="Phosphorus demand for soluble inerts"
        )
        def S_PD(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_F"]
                * blk.config.inlet_reaction_package.i_PSF
            )

        @self.Expression(
            self.flowsheet().time, doc="Organic nitrogen from soluble inerts"
        )
        def SN_org(blk, t):
            return blk.S_ND[t] / (blk.config.outlet_reaction_package.Ni["S_aa"] * mw_n)

        @self.Expression(self.flowsheet().time, doc="Monosaccharides mapping")
        def Ssu_mapping(blk, t):
            return Expr_if(
                blk.SN_org[t] >= blk.properties_in[t].conc_mass_comp["S_F"],
                eps * pyunits.kg / pyunits.m**3,
                blk.properties_in[t].conc_mass_comp["S_F"] - blk.SN_org[t],
            )

        @self.Expression(self.flowsheet().time, doc="Amino acids mapping")
        def Saa_mapping(blk, t):
            return Expr_if(
                blk.SN_org[t] >= blk.properties_in[t].conc_mass_comp["S_F"],
                blk.properties_in[t].conc_mass_comp["S_F"],
                blk.SN_org[t],
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Monosaccharides output",
        )
        def Ssu_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_su"] == blk.Ssu_mapping[t]

        @self.Constraint(
            self.flowsheet().time,
            doc="Amino acids output",
        )
        def Saa_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_aa"] == blk.Saa_mapping[t]

        @self.Expression(self.flowsheet().time, doc="S_F concentration at step 3")
        def SF_AS3(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_F"]
                - blk.Ssu_mapping[t]
                - blk.Saa_mapping[t]
            )

        @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 3")
        def SNH4_AS3(blk, t):
            return (
                blk.SNH4_AS2[t]
                + blk.properties_in[t].conc_mass_comp["S_F"]
                * blk.config.inlet_reaction_package.i_NSF
                - blk.Saa_mapping[t]
                * blk.config.outlet_reaction_package.Ni["S_aa"]
                * mw_n
            )

        @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 3")
        def SPO4_AS3(blk, t):
            return (
                blk.SPO4_AS2[t]
                + blk.properties_in[t].conc_mass_comp["S_F"]
                * blk.config.inlet_reaction_package.i_PSF
            )

        @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 3")
        def SIC_AS3(blk, t):
            return (
                blk.SIC_AS2[t]
                + blk.properties_in[t].conc_mass_comp["S_F"]
                * blk.config.inlet_reaction_package.i_CSF
                - blk.Ssu_mapping[t]
                * blk.config.outlet_reaction_package.Ci["S_su"]
                * mw_c
                - blk.Saa_mapping[t]
                * blk.config.outlet_reaction_package.Ci["S_aa"]
                * mw_c
            )

        # -----------------------------------------DecaySwitch.on----------------------------------------------------------#

        # -------------------------------------------Step 4----------------------------------------------------------------#
        if self.config.inlet_reaction_package.config.decay_switch == DecaySwitch.on:

            @self.Expression(
                self.flowsheet().time, doc="Biomass concentration at step 4"
            )
            # TODO: include sulfur (X_SO) in this expression after implementing sulfur-extension
            def biomass(blk, t):
                return (
                    blk.XH_AS2[t]
                    + blk.properties_in[t].conc_mass_comp["X_PAO"]
                    + blk.properties_in[t].conc_mass_comp["X_AUT"]
                )

            @self.Expression(self.flowsheet().time, doc="S_I concentration at step 4")
            def SI_AS4(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["S_I"]
                    + blk.biomass[t] * self.f_sI_xc
                )

            @self.Constraint(self.flowsheet().time, doc="S_I concentration output")
            def SI_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_I"] == blk.SI_AS4[t]

            @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 4")
            def SNH4_AS4(blk, t):
                return (
                    blk.SNH4_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_NBM
                    - blk.biomass[t]
                    * self.f_sI_xc
                    * blk.config.inlet_reaction_package.i_NSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_NSI
                    - blk.biomass[t]
                    * self.f_pr_xc
                    * blk.config.outlet_reaction_package.Ni["X_pr"]
                    * mw_n
                )

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 4")
            def SPO4_AS4(blk, t):
                return (
                    blk.SPO4_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_PBM
                    - blk.biomass[t] * self.f_sI_xc * self.i_PSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_PXI
                    - blk.biomass[t]
                    * self.f_ch_xc
                    * eps
                    * pyunits.kmol
                    / pyunits.kg
                    * mw_p
                    - blk.biomass[t]
                    * self.f_li_xc
                    * blk.config.outlet_reaction_package.Pi["X_li"]
                    * mw_p
                )

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 4")
            def SIC_AS4(blk, t):
                return (
                    blk.SIC_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_CXB
                    - blk.biomass[t]
                    * self.f_sI_xc
                    * blk.config.inlet_reaction_package.i_CSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_CXI
                    - blk.biomass[t]
                    * self.f_pr_xc
                    * blk.config.outlet_reaction_package.Ci["X_pr"]
                    * mw_c
                    - blk.biomass[t]
                    * self.f_ch_xc
                    * blk.config.outlet_reaction_package.Ci["X_ch"]
                    * mw_c
                    - blk.biomass[t]
                    * self.f_li_xc
                    * blk.config.outlet_reaction_package.Ci["X_li"]
                    * mw_c
                )

            @self.Expression(self.flowsheet().time, doc="X_I concentration at step 4")
            def XI_AS4(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_I"]
                    + blk.biomass[t] * self.f_xI_xc
                )

            @self.Constraint(self.flowsheet().time, doc="X_I concentration output")
            def XI_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_I"] == blk.XI_AS4[t]

            # TODO: Consider removing cases where vars are set to 0 and they are not being used
            # TODO: These were left in just to match Flores-Alsina as closely as possible
            XH_AS4 = eps * pyunits.kg / pyunits.m**3

            XPAO_AS4 = eps * pyunits.kg / pyunits.m**3

            @self.Constraint(self.flowsheet().time, doc="X_PAO concentration output")
            def XPAO_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PAO"] == XPAO_AS4

            XAUT_AS4 = eps * pyunits.kg / pyunits.m**3

            # -----------------------------------------Step 5--------------------------------------------------------------#
            @self.Expression(
                self.flowsheet().time, doc="Nitrogen demand for particulate inerts"
            )
            def X_ND(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_NXS
                )

            @self.Expression(
                self.flowsheet().time, doc="Phosphorus demand for particulate inerts"
            )
            def X_PD(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_PXS
                )

            @self.Expression(
                self.flowsheet().time, doc="Organic nitrogen from particulate inerts"
            )
            def XN_org(blk, t):
                return blk.X_ND[t] / (
                    blk.config.outlet_reaction_package.Ni["X_pr"] * mw_n
                )

            @self.Expression(self.flowsheet().time, doc="Carbohydrates mapping")
            def Xch_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    eps * pyunits.kg / pyunits.m**3,
                    (blk.properties_in[t].conc_mass_comp["X_S"] - blk.XN_org[t]) * 0.4,
                )

            @self.Expression(self.flowsheet().time, doc="Protein mapping")
            def Xpr_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    blk.SF_AS3[t],
                    blk.XN_org[t],
                )

            @self.Expression(self.flowsheet().time, doc="Lipids mapping")
            def Xli_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    eps * pyunits.kg / pyunits.m**3,
                    (blk.properties_in[t].conc_mass_comp["X_S"] - blk.XN_org[t]) * 0.6,
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Carbohydrates output",
            )
            def Xch_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_ch"]
                    == blk.Xch_mapping[t] + blk.biomass[t] * self.f_ch_xc
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Protein output",
            )
            def Xpr_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_pr"]
                    == blk.Xpr_mapping[t] + blk.biomass[t] * self.f_pr_xc
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Lipids output",
            )
            def Xli_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_li"]
                    == blk.Xli_mapping[t] + blk.biomass[t] * self.f_li_xc
                )

            @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 5")
            def SNH4_AS5(blk, t):
                return (
                    blk.SNH4_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_NXS
                    - blk.Xpr_mapping[t]
                    * blk.config.outlet_reaction_package.Ni["X_pr"]
                    * mw_n
                )

            @self.Constraint(self.flowsheet().time, doc="S_NH4 concentration output")
            def SIN_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IN"] == blk.SNH4_AS5[t]

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 5")
            def SPO4_AS5(blk, t):
                return (
                    blk.SPO4_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_PXS
                    - blk.Xch_mapping[t] * eps * pyunits.kmol / pyunits.kg * mw_p
                    - blk.Xli_mapping[t]
                    * blk.config.outlet_reaction_package.Pi["X_li"]
                    * mw_p
                )

            @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 5")
            def SIC_AS5(blk, t):
                return (
                    blk.SIC_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_CXS
                    - blk.Xch_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_ch"]
                    * mw_c
                    - blk.Xpr_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_pr"]
                    * mw_c
                    - blk.Xli_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_li"]
                    * mw_c
                )

            XS_AS5 = eps * pyunits.kg / pyunits.m**3

            # -----------------------------------------Step 6--------------------------------------------------------------#
            XPP_AS6 = eps * pyunits.kg / pyunits.m**3

            @self.Constraint(self.flowsheet().time, doc="X_PP concentration output")
            def XPP_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PP"] == XPP_AS6

            XPHA_AS6 = eps * pyunits.kg / pyunits.m**3

            @self.Constraint(self.flowsheet().time, doc="X_PHA concentration output")
            def XPHA_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PHA"] == XPHA_AS6

            Sva_AS6 = XPHA_AS6 * self.f_XPHA_Sva

            @self.Constraint(
                self.flowsheet().time,
                doc="Total valerate concentration output",
            )
            def Sva_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_va"] == Sva_AS6

            Sbu_AS6 = XPHA_AS6 * self.f_XPHA_Sbu

            @self.Constraint(
                self.flowsheet().time,
                doc="Total butyrate concentration output",
            )
            def Sbu_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_bu"] == Sbu_AS6

            Spro_AS6 = XPHA_AS6 * self.f_XPHA_Spro

            @self.Constraint(
                self.flowsheet().time,
                doc="Total propionate concentration output",
            )
            def Spro_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_pro"] == Spro_AS6

            Sac_AS6 = XPHA_AS6 * self.f_XPHA_Sac

            @self.Constraint(
                self.flowsheet().time,
                doc="Total acetate concentration output",
            )
            def Sac_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_ac"]
                    == Sac_AS6 + blk.SA_AS2[t]
                )

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 6")
            def SPO4_AS6(blk, t):
                return blk.SPO4_AS5[t] + blk.properties_in[t].conc_mass_comp["X_PP"]

            @self.Constraint(self.flowsheet().time, doc="S_IP concentration output")
            def SIP_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IP"] == blk.SPO4_AS6[t]

            @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 6")
            def SIC_AS6(blk, t):
                return (
                    blk.SIC_AS5[t]
                    + blk.properties_in[t].conc_mass_comp["X_PHA"] * self.C_PHA
                    - Sva_AS6 * blk.config.outlet_reaction_package.Ci["S_va"] * mw_c
                    - Sbu_AS6 * blk.config.outlet_reaction_package.Ci["S_bu"] * mw_c
                    - Spro_AS6 * blk.config.outlet_reaction_package.Ci["S_pro"] * mw_c
                    - Sac_AS6 * blk.config.outlet_reaction_package.Ci["S_ac"] * mw_c
                )

            @self.Constraint(self.flowsheet().time, doc="S_IC concentration output")
            def SIC_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IC"] == blk.SIC_AS6[t]

            @self.Expression(self.flowsheet().time, doc="S_K concentration at step 6")
            def SK_AS6(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["S_K"]
                    + blk.config.outlet_reaction_package.K_XPP
                    * blk.properties_in[t].conc_mass_comp["X_PP"]
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total potassium concentration output",
            )
            def SK_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_K"] == blk.SK_AS6[t]

            @self.Expression(self.flowsheet().time, doc="S_Mg concentration at step 6")
            def SMg_AS6(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["S_Mg"]
                    + blk.config.outlet_reaction_package.Mg_XPP
                    * blk.properties_in[t].conc_mass_comp["X_PP"]
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total magnesium concentration output",
            )
            def SMg_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_Mg"] == blk.SMg_AS6[t]

        # ---------------------------------------DecaySwitch.off-------------------------------------------------------#

        # -----------------------------------------Step 4--------------------------------------------------------------#
        elif self.config.inlet_reaction_package.config.decay_switch == DecaySwitch.off:

            @self.Expression(
                self.flowsheet().time, doc="Biomass concentration at step 4"
            )
            # TODO: include sulfur (X_SO) in this expression after implementing sulfur-extension
            def biomass(blk, t):
                return blk.XH_AS2[t] + blk.properties_in[t].conc_mass_comp["X_AUT"]

            @self.Expression(self.flowsheet().time, doc="S_I concentration at step 4")
            def SI_AS4(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["S_I"]
                    + blk.biomass[t] * self.f_sI_xc
                )

            @self.Constraint(self.flowsheet().time, doc="S_I concentration output")
            def SI_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_I"] == blk.SI_AS4[t]

            @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 4")
            def SNH4_AS4(blk, t):
                return (
                    blk.SNH4_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_NBM
                    - blk.biomass[t]
                    * self.f_sI_xc
                    * blk.config.inlet_reaction_package.i_NSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_NSI
                    - blk.biomass[t]
                    * self.f_pr_xc
                    * blk.config.outlet_reaction_package.Ni["X_pr"]
                    * mw_n
                )

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 4")
            def SPO4_AS4(blk, t):
                return (
                    blk.SPO4_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_PBM
                    - blk.biomass[t] * self.f_sI_xc * self.i_PSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_PXI
                    - blk.biomass[t]
                    * self.f_ch_xc
                    * eps
                    * pyunits.kmol
                    / pyunits.kg
                    * mw_p
                    - blk.biomass[t]
                    * self.f_li_xc
                    * blk.config.outlet_reaction_package.Pi["X_li"]
                    * mw_p
                )

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 4")
            def SIC_AS4(blk, t):
                return (
                    blk.SIC_AS3[t]
                    + blk.biomass[t] * blk.config.inlet_reaction_package.i_CXB
                    - blk.biomass[t]
                    * self.f_sI_xc
                    * blk.config.inlet_reaction_package.i_CSI
                    - blk.biomass[t]
                    * self.f_xI_xc
                    * blk.config.inlet_reaction_package.i_CXI
                    - blk.biomass[t]
                    * self.f_pr_xc
                    * blk.config.outlet_reaction_package.Ci["X_pr"]
                    * mw_c
                    - blk.biomass[t]
                    * self.f_ch_xc
                    * blk.config.outlet_reaction_package.Ci["X_ch"]
                    * mw_c
                    - blk.biomass[t]
                    * self.f_li_xc
                    * blk.config.outlet_reaction_package.Ci["X_li"]
                    * mw_c
                )

            @self.Expression(self.flowsheet().time, doc="X_I concentration at step 4")
            def XI_AS4(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_I"]
                    + blk.biomass[t] * self.f_xI_xc
                )

            @self.Constraint(self.flowsheet().time, doc="X_I concentration output")
            def XI_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_I"] == blk.XI_AS4[t]

            XH_AS4 = eps * pyunits.kg / pyunits.m**3

            @self.Expression(self.flowsheet().time, doc="X_PAO concentration at step 4")
            def XPAO_AS4(blk, t):
                return blk.properties_in[t].conc_mass_comp["X_PAO"]

            @self.Expression(self.flowsheet().time, doc="X_PP concentration at step 4")
            def XPP_AS4(blk, t):
                return blk.properties_in[t].conc_mass_comp["X_PP"]

            @self.Expression(self.flowsheet().time, doc="X_PHA concentration at step 4")
            def XPHA_AS4(blk, t):
                return blk.properties_in[t].conc_mass_comp["X_PHA"]

            XAUT_AS4 = eps * pyunits.kg / pyunits.m**3

            # -----------------------------------------Step 5--------------------------------------------------------------#
            @self.Expression(
                self.flowsheet().time, doc="Nitrogen demand for particulate inerts"
            )
            def X_ND(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_NXS
                )

            @self.Expression(
                self.flowsheet().time, doc="Phosphorus demand for particulate inerts"
            )
            def X_PD(blk, t):
                return (
                    blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_PXS
                )

            @self.Expression(
                self.flowsheet().time, doc="Organic nitrogen from particulate inerts"
            )
            def XN_org(blk, t):
                return blk.X_ND[t] / (
                    blk.config.outlet_reaction_package.Ni["X_pr"] * mw_n
                )

            @self.Expression(self.flowsheet().time, doc="Carbohydrates mapping")
            def Xch_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    eps * pyunits.kg / pyunits.m**3,
                    (blk.properties_in[t].conc_mass_comp["X_S"] - blk.XN_org[t]) * 0.4,
                )

            @self.Expression(self.flowsheet().time, doc="Protein mapping")
            def Xpr_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    blk.SF_AS3[t],
                    blk.XN_org[t],
                )

            @self.Expression(self.flowsheet().time, doc="Lipids mapping")
            def Xli_mapping(blk, t):
                return Expr_if(
                    blk.XN_org[t] >= blk.properties_in[t].conc_mass_comp["X_S"],
                    eps * pyunits.kg / pyunits.m**3,
                    (blk.properties_in[t].conc_mass_comp["X_S"] - blk.XN_org[t]) * 0.6,
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Carbohydrates output",
            )
            def Xch_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_ch"]
                    == blk.Xch_mapping[t] + blk.biomass[t] * self.f_ch_xc
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Protein output",
            )
            def Xpr_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_pr"]
                    == blk.Xpr_mapping[t] + blk.biomass[t] * self.f_pr_xc
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Lipids output",
            )
            def Xli_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["X_li"]
                    == blk.Xli_mapping[t] + blk.biomass[t] * self.f_li_xc
                )

            @self.Expression(self.flowsheet().time, doc="S_NH4 concentration at step 5")
            def SNH4_AS5(blk, t):
                return (
                    blk.SNH4_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_NXS
                    - blk.Xpr_mapping[t]
                    * blk.config.outlet_reaction_package.Ni["X_pr"]
                    * mw_n
                )

            @self.Constraint(self.flowsheet().time, doc="S_NH4 concentration output")
            def SIN_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IN"] == blk.SNH4_AS5[t]

            @self.Expression(self.flowsheet().time, doc="S_PO4 concentration at step 5")
            def SPO4_AS5(blk, t):
                return (
                    blk.SPO4_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_PXS
                    - blk.Xch_mapping[t] * eps * pyunits.kmol / pyunits.kg * mw_p
                    - blk.Xli_mapping[t]
                    * blk.config.outlet_reaction_package.Pi["X_li"]
                    * mw_p
                )

            @self.Constraint(self.flowsheet().time, doc="S_IP concentration output")
            def SIP_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IP"] == blk.SPO4_AS5[t]

            @self.Expression(self.flowsheet().time, doc="S_IC concentration at step 5")
            def SIC_AS5(blk, t):
                return (
                    blk.SIC_AS4[t]
                    + blk.properties_in[t].conc_mass_comp["X_S"]
                    * blk.config.inlet_reaction_package.i_CXS
                    - blk.Xch_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_ch"]
                    * mw_c
                    - blk.Xpr_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_pr"]
                    * mw_c
                    - blk.Xli_mapping[t]
                    * blk.config.outlet_reaction_package.Ci["X_li"]
                    * mw_c
                )

            @self.Constraint(self.flowsheet().time, doc="S_IC concentration output")
            def SIC_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["S_IC"] == blk.SIC_AS5[t]

            XS_AS5 = eps * pyunits.kg / pyunits.m**3

            XPAO_AS5 = self.properties_in[0].conc_mass_comp["X_PAO"]

            XPP_AS5 = self.properties_in[0].conc_mass_comp["X_PP"]

            XPHA_AS5 = self.properties_in[0].conc_mass_comp["X_PHA"]

            # -----------------------------------------Step 6--------------------------------------------------------------#
            @self.Constraint(
                self.flowsheet().time,
                doc="Total valerate concentration",
            )
            def Sva_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_va"]
                    == eps * pyunits.kg / pyunits.m**3
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total butyrate concentration",
            )
            def Sbu_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_bu"]
                    == eps * pyunits.kg / pyunits.m**3
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total propionate concentration",
            )
            def Spro_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_pro"]
                    == eps * pyunits.kg / pyunits.m**3
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total acetate concentration",
            )
            def Sac_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_ac"]
                    == eps * pyunits.kg / pyunits.m**3 + blk.SA_AS2[t]
                )

            XPAO_AS6 = self.properties_in[0].conc_mass_comp["X_PAO"]

            @self.Constraint(self.flowsheet().time, doc="X_PAO concentration output")
            def XPAO_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PAO"] == XPAO_AS6

            XPP_AS6 = self.properties_in[0].conc_mass_comp["X_PP"]

            @self.Constraint(self.flowsheet().time, doc="X_PP concentration output")
            def XPP_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PP"] == XPP_AS6

            XPHA_AS6 = self.properties_in[0].conc_mass_comp["X_PHA"]

            @self.Constraint(self.flowsheet().time, doc="X_PHA concentration output")
            def XPHA_output(blk, t):
                return blk.properties_out[t].conc_mass_comp["X_PHA"] == XPHA_AS6

            @self.Constraint(
                self.flowsheet().time,
                doc="Total magnesium concentration output",
            )
            def SK_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_K"]
                    == blk.properties_in[t].conc_mass_comp["S_K"]
                )

            @self.Constraint(
                self.flowsheet().time,
                doc="Total magnesium concentration output",
            )
            def SMg_output(blk, t):
                return (
                    blk.properties_out[t].conc_mass_comp["S_Mg"]
                    == blk.properties_in[t].conc_mass_comp["S_Mg"]
                )

        self.zero_flow_components = Set(
            initialize=[
                "S_fa",
                "S_h2",
                "S_ch4",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
            ]
        )

        @self.Constraint(
            self.flowsheet().time,
            self.zero_flow_components,
            doc="Components with no flow equation",
        )
        def return_zero_flow_comp(blk, t, i):
            return (
                blk.properties_out[t].conc_mass_comp[i]
                == eps * pyunits.kg / pyunits.m**3
            )

        # TODO: Relationship carried over from ASM1/ADM1 - not clear how Flores-Alsina handles ions (see perf_plant_AD_ss.m?)
        @self.Constraint(
            self.flowsheet().time,
            doc="Anions balance",
        )
        def return_an(blk, t):
            return (
                blk.properties_out[t].anions
                == blk.properties_out[t].conc_mass_comp["S_IN"] / mw_n
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Cations balance",
        )
        def return_cat(blk, t):
            return (
                blk.properties_out[t].cations
                == blk.properties_out[t].conc_mass_comp["S_IC"] / mw_c
            )

    def initialize_build(
        self,
        state_args_in=None,
        state_args_out=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        This method calls the initialization method of the state blocks.

        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            state_args_out : a dict of arguments to be passed to the outlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state block
        flags = self.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_in,
            hold_state=True,
        )

        self.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        if degrees_of_freedom(self) != 0:
            raise Exception(
                f"{self.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(self)}"
            )

        with idaeslog.solver_log(init_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        self.properties_in.release_state(flags=flags, outlvl=outlvl)

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import contextlib
import enum
from pathlib import Path
from typing import Container, Optional, Callable

import pytest
from _pytest.nodes import Item
from _pytest.config import Config
from _pytest.config.argparsing import Parser


class MarkerSpec(enum.Enum):
    unit = "Quick tests that do not require a solver, must run in < 2 s"
    component = "Quick tests that may require a solver"
    integration = "Long duration tests"
    build = "FIXME for building stuff?"
    solver = "Tests that require a solver"
    requires_idaes_solver = (
        "Tests that require a solver from the IDEAS extensions to pass"
    )

    @property
    def description(self) -> str:
        return self.value

    @classmethod
    def for_item(cls, item: Item) -> Container["MarkerSpec"]:
        found = []
        for marker in item.iter_markers():
            with contextlib.suppress(KeyError):
                found.append(cls[marker.name])
        return found


def _handle_requires_idaes_solver(
    solver: Optional = None, action: Optional[Callable[[str], None]] = pytest.xfail
) -> None:
    from idaes.core.solvers import get_solver
    from idaes.config import bin_directory

    solver = solver or get_solver()
    idaes_bin_dir = Path(bin_directory).resolve()
    solver_bin_path = Path(solver.executable()).resolve()

    if not idaes_bin_dir in solver_bin_path.parents:
        action(f"This test is known to be failing with {solver_bin_path}")


def pytest_configure(config: Config):

    for marker_spec in MarkerSpec:
        config.addinivalue_line(
            "markers", f"{marker_spec.name}: {marker_spec.description}"
        )


def pytest_runtest_setup(item: Item):

    if MarkerSpec.requires_idaes_solver in MarkerSpec.for_item(item):
        # TODO we could get some more information about a specific solver,
        # either by providing args to the marker
        # or by inspecting the current value of the `solver` fixture
        _handle_requires_idaes_solver()


def pytest_addoption(parser: Parser):
    parser.addoption(
        "--edb-no-mock",
        help="Force the `edb` fixture to connect to a running MongoDB instance "
        "instead of falling back to mongomock",
        action="store_true",
        default=False,
        dest="edb_no_mock",
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    TerminationCondition,
    SolverStatus,
    value,
    Constraint,
    Var,
)

from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
    unscaled_variables_generator,
    unscaled_constraints_generator,
    get_scaling_factor,
    set_scaling_factor,
)

import watertap.property_models.seawater_prop_pack as props
import watertap.property_models.multicomp_aq_sol_prop_pack as props2
from watertap.unit_models.pressure_changer import (
    Pump,
    EnergyRecoveryDevice,
    VariableEfficiency,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------


class TestPumpIsothermal:
    @pytest.fixture(scope="class")
    def Pump_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.SeawaterParameterBlock()

        m.fs.unit = Pump(property_package=m.fs.properties)

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_TDS = 0.035
        feed_pressure_in = 1e5
        feed_pressure_out = 5e5
        feed_temperature = 273.15 + 25
        efi_pump = 0.75

        feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
            feed_flow_mass * feed_mass_frac_TDS
        )
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mass * feed_mass_frac_H2O
        )
        m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
        m.fs.unit.efficiency_pump.fix(efi_pump)
        return m

    @pytest.mark.unit
    def test_build(self, Pump_frame):
        m = Pump_frame

        # check that IDAES pump model has not changed variable and constraint names
        var_list = [
            "work_mechanical",
            "deltaP",
            "ratioP",
            "work_fluid",
            "efficiency_pump",
        ]
        for v in var_list:
            assert hasattr(m.fs.unit, v)
            var = getattr(m.fs.unit, v)
            assert isinstance(var, Var)

        con_list = ["ratioP_calculation", "fluid_work_calculation", "actual_work"]
        for c in con_list:
            assert hasattr(m.fs.unit, c)
            con = getattr(m.fs.unit, c)
            assert isinstance(con, Constraint)

        # check that IDAES control volume has not changed variable and constraint names
        assert hasattr(m.fs.unit.control_volume, "work")
        assert hasattr(m.fs.unit.control_volume, "deltaP")

        assert hasattr(m.fs.unit.control_volume, "material_balances")
        assert hasattr(m.fs.unit.control_volume, "pressure_balance")

        # check that energy balance was removed
        assert not hasattr(m.fs.unit.control_volume, "enthalpy_balances")

        # check that isothermal constraint was added
        assert hasattr(m.fs.unit.control_volume, "isothermal_balance")
        assert isinstance(m.fs.unit.control_volume.isothermal_balance, Constraint)

    @pytest.mark.unit
    def test_dof(self, Pump_frame):
        m = Pump_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, Pump_frame):
        m = Pump_frame

        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )
        calculate_scaling_factors(m)

        if (
            get_scaling_factor(m.fs.unit.ratioP) is None
        ):  # if IDAES hasn't specified a scaling factor
            set_scaling_factor(m.fs.unit.ratioP, 1)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, Pump_frame):
        initialization_tester(Pump_frame)

    @pytest.mark.component
    def test_solve(self, Pump_frame):
        m = Pump_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_conservation(self, Pump_frame):
        m = Pump_frame
        b = m.fs.unit.control_volume
        comp_lst = ["TDS", "H2O"]

        for t in m.fs.config.time:
            # mass balance
            for j in comp_lst:
                assert (
                    abs(
                        value(
                            b.properties_in[t].flow_mass_phase_comp["Liq", j]
                            - b.properties_out[t].flow_mass_phase_comp["Liq", j]
                        )
                    )
                    <= 1e-6
                )

    @pytest.mark.component
    def test_solution(self, Pump_frame):
        m = Pump_frame

        # pump variables
        assert pytest.approx(521.06, rel=1e-3) == value(m.fs.unit.work_mechanical[0])
        assert pytest.approx(4e5, rel=1e-3) == value(m.fs.unit.deltaP[0])
        assert pytest.approx(5, rel=1e-3) == value(m.fs.unit.ratioP[0])
        assert pytest.approx(390.79, rel=1e-3) == value(m.fs.unit.work_fluid[0])

        # outlet state variables
        assert pytest.approx(0.965, rel=1e-3) == value(
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        )
        assert pytest.approx(0.035, rel=1e-3) == value(
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        )
        assert pytest.approx(298.15, rel=1e-5) == value(
            m.fs.unit.control_volume.properties_out[0].temperature
        )
        assert pytest.approx(9.9318e4, rel=1e-5) == value(
            m.fs.unit.control_volume.properties_out[0].enth_flow
        )


class TestEnergyRecoveryDevice(TestPumpIsothermal):
    @pytest.fixture(scope="class")
    def Pump_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.SeawaterParameterBlock()

        m.fs.unit = EnergyRecoveryDevice(property_package=m.fs.properties)

        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_TDS = 0.035
        feed_pressure_in = 5e5
        feed_pressure_out = 1e5
        feed_temperature = 273.15 + 25
        efi_pump = 0.75

        feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
            feed_flow_mass * feed_mass_frac_TDS
        )
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mass * feed_mass_frac_H2O
        )
        m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
        m.fs.unit.efficiency_pump.fix(efi_pump)
        return m

    @pytest.mark.component
    def test_solution(self, Pump_frame):
        m = Pump_frame

        # pump variables
        assert pytest.approx(-293.09, rel=1e-3) == value(m.fs.unit.work_mechanical[0])
        assert pytest.approx(-4e5, rel=1e-3) == value(m.fs.unit.deltaP[0])
        assert pytest.approx(1.0 / 5, rel=1e-3) == value(m.fs.unit.ratioP[0])
        assert pytest.approx(-390.79, rel=1e-3) == value(m.fs.unit.work_fluid[0])

        # outlet state variables
        assert pytest.approx(0.965, rel=1e-3) == value(
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        )
        assert pytest.approx(0.035, rel=1e-3) == value(
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        )
        assert pytest.approx(298.15, rel=1e-5) == value(
            m.fs.unit.control_volume.properties_out[0].temperature
        )


class TestPumpVariable_Flow:
    @pytest.fixture(scope="class")
    def Pump_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.SeawaterParameterBlock()

        m.fs.unit = Pump(
            property_package=m.fs.properties,
            variable_efficiency=VariableEfficiency.flow,
        )
        # fully specify system
        feed_flow_mass = 1
        feed_mass_frac_TDS = 0.035
        feed_pressure_in = 1e5
        feed_pressure_out = 5e5
        feed_temperature = 273.15 + 25
        efi_pump = 0.75

        feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
            feed_flow_mass * feed_mass_frac_TDS
        )
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mass * feed_mass_frac_H2O
        )
        m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)

        m.fs.unit.bep_eta.fix(efi_pump)
        m.fs.unit.flow_ratio.fix(1)
        return m

    @pytest.mark.unit
    def test_build(self, Pump_frame):
        m = Pump_frame

        # check that IDAES pump model has not changed variable and constraint names
        var_list = [
            "bep_flow",
            "bep_eta",
            "flow_ratio",
        ]
        for v in var_list:
            assert hasattr(m.fs.unit, v)
            var = getattr(m.fs.unit, v)
            assert isinstance(var, Var)

        assert isinstance(m.fs.unit.flow_ratio_constraint, Constraint)

    @pytest.mark.unit
    def test_calculate_scaling(self, Pump_frame):
        m = Pump_frame
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )

        calculate_scaling_factors(m)

        # if IDAES hasn't specified a scaling factor
        if get_scaling_factor(m.fs.unit.ratioP) is None:
            set_scaling_factor(m.fs.unit.ratioP, 1)

        if get_scaling_factor(m.fs.unit.control_volume.work) is None:
            set_scaling_factor(m.fs.unit.control_volume.work, 1)

        # check that all variables and constraints have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0

        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.unit
    def test_dof(self, Pump_frame):
        m = Pump_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    def test_initialize(self, Pump_frame):
        initialization_tester(Pump_frame)

    @pytest.mark.component
    def test_solve(self, Pump_frame):
        results = solver.solve(Pump_frame)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.unit
    def test_conservation(self, Pump_frame):
        m = Pump_frame
        b = m.fs.unit.control_volume
        comp_lst = ["TDS", "H2O"]

        for t in m.fs.config.time:
            # mass balance
            for j in comp_lst:
                assert (
                    abs(
                        value(
                            b.properties_in[t].flow_mass_phase_comp["Liq", j]
                            - b.properties_out[t].flow_mass_phase_comp["Liq", j]
                        )
                    )
                    <= 1e-6
                )

    @pytest.mark.unit
    def test_units(self, Pump_frame):
        assert_units_consistent(Pump_frame)

    @pytest.mark.component
    def test_solution(self, Pump_frame):
        m = Pump_frame

        default_flow_vol = m.fs.unit.bep_flow()
        assert pytest.approx(9.9318e4, rel=1e-5) == value(
            m.fs.unit.control_volume.properties_out[0].enth_flow
        )
        assert pytest.approx(1, rel=1e-3) == value(m.fs.unit.eta_ratio[0])

        # Test low bep flow case
        m.fs.unit.flow_ratio.unfix()
        m.fs.unit.bep_flow.fix(default_flow_vol / 2)
        results = solver.solve(m)
        assert pytest.approx(0.4, rel=1e-5) == value(m.fs.unit.eta_ratio[0])

        # Test high bep flow case
        m.fs.unit.bep_flow.fix(default_flow_vol * 2)
        results = solver.solve(m)
        assert pytest.approx(0.4, rel=1e-5) == value(m.fs.unit.eta_ratio[0])

        # Test sub-optimal bep flow case
        m.fs.unit.bep_flow.fix(default_flow_vol * 0.8)
        results = solver.solve(m)
        assert pytest.approx(0.9345625, rel=1e-5) == value(m.fs.unit.eta_ratio[0])


class TestPumpIsothermal_with_energybalancetype_none:
    @pytest.fixture(scope="class")
    def Pump_frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props2.MCASParameterBlock(
            solute_list=["TDS"],
            mw_data={"TDS": 58e-3},
            ignore_neutral_charge=True,
        )

        m.fs.unit = Pump(property_package=m.fs.properties)

        # fully specify system
        feed_flow_mol = 1
        feed_mol_frac_TDS = 0.035
        feed_pressure_in = 1e5
        feed_pressure_out = 5e5
        feed_temperature = 273.15 + 25
        efi_pump = 0.75

        feed_mass_frac_H2O = 1 - feed_mol_frac_TDS
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "TDS"].fix(
            feed_flow_mol * feed_mol_frac_TDS
        )
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mol * feed_mass_frac_H2O
        )
        m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
        m.fs.unit.inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
        m.fs.unit.efficiency_pump.fix(efi_pump)
        return m

    @pytest.mark.unit
    def test_build(self, Pump_frame):
        m = Pump_frame

        # check that IDAES pump model has not changed variable and constraint names
        var_list = [
            "work_mechanical",
            "deltaP",
            "ratioP",
            "work_fluid",
            "efficiency_pump",
        ]
        for v in var_list:
            assert hasattr(m.fs.unit, v)
            var = getattr(m.fs.unit, v)
            assert isinstance(var, Var)

        con_list = ["ratioP_calculation", "fluid_work_calculation", "actual_work"]
        for c in con_list:
            assert hasattr(m.fs.unit, c)
            con = getattr(m.fs.unit, c)
            assert isinstance(con, Constraint)

        # check that IDAES control volume has not changed variable and constraint names
        assert hasattr(m.fs.unit.control_volume, "work")
        assert hasattr(m.fs.unit.control_volume, "deltaP")

        assert hasattr(m.fs.unit.control_volume, "material_balances")
        assert hasattr(m.fs.unit.control_volume, "pressure_balance")
        assert not hasattr(m.fs.unit.control_volume, "energy_balance")

        # check that energy balance was removed
        assert not hasattr(m.fs.unit.control_volume, "enthalpy_balances")

        # check that isothermal constraint was added
        assert hasattr(m.fs.unit.control_volume, "isothermal_balance")
        assert isinstance(m.fs.unit.control_volume.isothermal_balance, Constraint)

    @pytest.mark.unit
    def test_dof(self, Pump_frame):
        m = Pump_frame
        assert degrees_of_freedom(m) == 0

    @pytest.mark.unit
    def test_calculate_scaling(self, Pump_frame):
        m = Pump_frame

        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mol_phase_comp", 1e2, index=("Liq", "TDS")
        )
        calculate_scaling_factors(m)

        if (
            get_scaling_factor(m.fs.unit.ratioP) is None
        ):  # if IDAES hasn't specified a scaling factor
            set_scaling_factor(m.fs.unit.ratioP, 1)

        # check that all variables have scaling factors
        unscaled_var_list = list(unscaled_variables_generator(m))
        assert len(unscaled_var_list) == 0
        # check that all constraints have been scaled
        unscaled_constraint_list = list(unscaled_constraints_generator(m))
        assert len(unscaled_constraint_list) == 0

    @pytest.mark.component
    def test_initialize(self, Pump_frame):
        initialization_tester(Pump_frame)

    @pytest.mark.component
    def test_solve(self, Pump_frame):
        m = Pump_frame
        results = solver.solve(m)

        # Check for optimal solution
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_conservation(self, Pump_frame):
        m = Pump_frame
        b = m.fs.unit.control_volume
        comp_lst = ["TDS", "H2O"]

        for t in m.fs.config.time:
            # mole balance
            for j in comp_lst:
                assert (
                    abs(
                        value(
                            b.properties_in[t].flow_mol_phase_comp["Liq", j]
                            - b.properties_out[t].flow_mol_phase_comp["Liq", j]
                        )
                    )
                    <= 1e-6
                )
